package org.cybertory.MSS;

/*-------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
/**
 *  Provides a bias in the calculation of radius in a particular direction
 *  for vertices of a randomly generated polygon.
 *  See {@link CircleFeatureShapeFactory} and {@link PolygonFactory}.
 */
public class AngleBias {
    final private float angle;
    final private float sweep;
    final private float factor;


    /**
     * @param angle  the general direction of vertices to be affected 
     *               by the bias.
     * @param sweep  the range of angles affected by the bias.
     * @param factor the number to multiply by the radius (modified
     *               by how far from the center of the sweep).
     */
    public AngleBias(float angle, float sweep, float factor) {
        this.angle = angle;
        this.sweep = sweep;
        this.factor = factor;
    }


    /**
     *  Calculate the number to be multiplied by the vertex radius, given the
     *  angle of the vertex.
     *
     *  @param vertexAngle  the angle of the vertex
     *  @return  the number to be multiplied.
     */
    public float getMultiple(float vertexAngle) {
        float angleDiff = vertexAngle - angle;
        
        if (angleDiff < -(float)Math.PI) {
            angleDiff += (float)Math.PI*2f;
        }
        else if (angleDiff > (float)Math.PI) {
            angleDiff -= (float)Math.PI*2f;
        }

        return 1f + ((Math.abs(angleDiff) > sweep/2f) ? 0f :
                factor * (float)Math.cos( angleDiff * (float)Math.PI / sweep ));
    }


    /**
     *  Calculate the number to be multiplied by the vertex radius, given the
     *  angle of the vertex and a list of angle biases.
     *
     *  @param biases  a list of angle biases
     *  @param vertexAngle  the angle of the vertex
     *  @return  the number to be multiplied.
     */
    static public float getMultiple(AngleBias[] biases, float vertexAngle) {
        float multiple = 1f;
        if (biases != null) {
            for (int i=0; i<biases.length; ++i) {
                multiple *= biases[i].getMultiple(vertexAngle);
            }
        }

        return multiple;
    }
}
