package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Ellipse2D;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.Random;

/**
 *  A flavor of FeatureShapeFactory that produces circular features,
 *  including support for the variability of location, size, and shape.
 *  Relies heavily on the {@link PolygonFactory}.
 *
 *  Note that all random effects in this class are related to array 
 *  manufacture -- therefore in the Cybertory, the random seed should be
 *  based on the serial number.
 */

public class CircleFeatureShapeFactory  implements FeatureShapeFactory {

    private Random rand = null;
    private float locationVariability = 0.0f;
    private float sizeVariability = 0.0f;
    private int shapeVariability = 0;
    private boolean roundedCorners = true;

    private PolygonFactory polyFactory = null;

    
    /**
     *  Constructor; variability can be added by {@link #setVariability}.
     */
    public CircleFeatureShapeFactory() {
    }


    /**
     *  Generate a specific shape, given the design location of a feature.
     *
     *  @param featurePixelSize  size of feature in pixel units
     *  @param featureCenter  the design location of a feature; this could
     *                          by altered if setVariability is called.
     *  @return the {@link java.awt.Shape} of the feature.
     */
    public Shape createFeatureShape(float featurePixelSize, 
            Point2D featureCenter) {
        Shape fShape = createFeatureShape(featurePixelSize);

        Point2D fCenter = (locationVariability > 0.0f) ?
                perturbatedLocation(featureCenter) : featureCenter;

        AffineTransform translate = AffineTransform.getTranslateInstance(
                fCenter.getX(), fCenter.getY());

        return translate.createTransformedShape(fShape);
    }


    /**
     *  Add variability factors to the shape generation.
     *  Default (if this method isn't called) is for zero variability
     *  (perfect in place circles of uniform size).
     *
     *  @param rand  random number object (which caller has seeded)
     *  @param locationVariability Standard deviation of location from
     *        design location using a random angle.
     *  @param sizeVariability  This number is used as a base for an 
     *        exponential with random number as exponent, the result is 
     *        multiplied by the design size.
     *  @param shapeVariability  0=none, 1=low, 2=medium, 3=high
     *  @param roundedCorners  
     */
    public void setVariability(
            Random rand,
            float locationVariability,
            float sizeVariability,
            int shapeVariability,
            boolean roundedCorners,
            AngleBias[] biases) {

        this.rand = (rand != null) ? rand : new Random();
        this.locationVariability = locationVariability;
        this.sizeVariability = sizeVariability;
        this.shapeVariability = shapeVariability;
        this.roundedCorners = roundedCorners;

        if (shapeVariability > 0f) {
            polyFactory = new PolygonFactory(rand, shapeVariability);
            if (biases != null) {
                for (int i=0; i<biases.length; ++i) {
                    polyFactory.addAngleBias(biases[i]);
                }
            }
        }
    }

    
    /**
     *  Generate a specific shape, centered at 0,0.
     *
     *  @param featurePixelSize  size of feature in pixel units
     *  @return the {@link java.awt.Shape} of the feature.
     */
    public Shape createFeatureShape(float featurePixelSize) {
        float size = featurePixelSize;
        if (sizeVariability != 0.0f) {
            size *= Math.pow(sizeVariability, rand.nextGaussian());
        }

        Shape fShape = null;
        if (shapeVariability == 0.0f) {
            fShape = (Shape)
                   new Ellipse2D.Float(-0.5f * size, -0.5f * size, size, size);
        }

        else if (roundedCorners) {
            PolarCoords[] pCoords = polyFactory.createPolygonPolar(size/2f);
            GeneralPath gp = new GeneralPath();
            gp.moveTo((float)Math.cos(pCoords[0].angle)*pCoords[0].radius,
                      (float)Math.sin(pCoords[0].angle)*pCoords[0].radius);

            for (int i=0; i<pCoords.length; ++i) {
                float angle0 = pCoords[i].angle;
                float rad0 = pCoords[i].radius;
                float angle1 = (i==pCoords.length-1) ? pCoords[0].angle :
                        pCoords[i+1].angle;
                float rad1 = (i==pCoords.length-1) ? pCoords[0].radius :
                        pCoords[i+1].radius;

                float x0 = (float)Math.cos(angle0)*rad0;
                float y0 = (float)Math.sin(angle0)*rad0;

                float xCP0 = x0 + (float)Math.cos(angle0+Math.PI/2f)*rad0/5f;
                float yCP0 = y0 + (float)Math.sin(angle0+Math.PI/2f)*rad0/5f;

                float x1 = (float)Math.cos(angle1)*rad1;
                float y1 = (float)Math.sin(angle1)*rad1;

                float xCP1 = x1 + (float)Math.cos(angle1-Math.PI/2f)*rad1/5f;
                float yCP1 = y1 + (float)Math.sin(angle1-Math.PI/2f)*rad1/5f;

                gp.curveTo(xCP0, yCP0, xCP1, yCP1, x1, y1);
            }
            fShape = (Shape)gp;
        }

        else {
            fShape = (Shape)polyFactory.createPolygon(size/2f);
        }

        return fShape;
    }


    private Point2D perturbatedLocation(Point2D originalLocation) {
        double radius = rand.nextGaussian() * locationVariability;
        double angle = rand.nextFloat() * Math.PI;

        return new Point2D.Float( 
                (float)(originalLocation.getX() + radius * Math.cos(angle)),
                (float)(originalLocation.getY() + radius * Math.sin(angle))
                );
    }
}
