package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.RenderingHints;
import java.awt.image.DataBuffer;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.awt.image.renderable.RenderedImageFactory;

import javax.media.jai.BorderExtender;
import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.OperationDescriptor;
import javax.media.jai.OperationDescriptorImpl;
import javax.media.jai.OperationRegistry;
import javax.media.jai.ParameterListDescriptor;
import javax.media.jai.ParameterListDescriptorImpl;
import javax.media.jai.RenderedOp;
import javax.media.jai.registry.RIFRegistry;


/**
 * This class is designed to be useable in with
 * {@link PendingRenderedOp} /  {@link PendingParameterBlockJAI}.
 * It's a way of getting images with integer pixel values into the
 * float image domain.
 */
public class ConvertToFloatOpDesc  extends OperationDescriptorImpl
        implements RenderedImageFactory  {

    static public String OPERATION_NAME = "ConvertToFloat";

    static public final int SizeToFit = 1;
    static public final int PreserveAspectRatio = 2;
    static public final int PreserveSize = 3;

    private static boolean isInitialized = false;

    public static void init() {
        if (isInitialized)
            return;

        isInitialized = true;
        OperationRegistry opReg =
                JAI.getDefaultInstance().getOperationRegistry();
        OperationDescriptor od = new ConvertToFloatOpDesc();

        opReg.registerDescriptor(od);
        RIFRegistry.register(opReg, od.getName(), OPERATION_NAME,
                             (RenderedImageFactory)od);
    }


    public ConvertToFloatOpDesc( ) {

        super(
                // resources
                new String[][] {
                     { "GlobalName",  OPERATION_NAME },
                     { "LocalName",  OPERATION_NAME },
                     { "Vendor",  "Attotron Biosensor Corporation" },
                     { "Description",
                       "Convert size and data model to match a float image" },
                     { "DocURL",  "" },
                     { "Version",  "0.9" },
                     { "arg0Desc", "original image" },
                     { "arg1Desc", "Intensity Divisor" },
                     { "arg2Desc", "Image Layout" },
                     { "arg3Desc", "Preserve Aspect Ratio ?" },
                     { "arg4Desc", "Source diagonal size" },
                     { "arg5Desc", "Border extender" }
                },

                // supported modes
                new String[] { "Rendered" },

                // source names
                null,   // use default naming convention "source0", ...

                // source classes
                new Class[][] { { RenderedImage.class } },

                new ParameterListDescriptorImpl(
                    // descriptor ??
                    new Object(),

                    // paramNames
                    new String[] {
                            "originalImage",
                            "intensityDivisor",
                            "fImageLayout",
                            "preserveAspectRatio",
                            "sourceDiagonalSize",
                            "borderExtender"  },

                    // paramClasses
                    new Class[] {
                            RenderedImage.class,
                            Float.class,
                            FImageLayout.class, 
                            Boolean.class,
                            Float.class,
                            BorderExtender.class },

                    // paramDefaults
                    new Object[] {
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            null,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            Boolean.FALSE,
                            null,
                 BorderExtender.createInstance(BorderExtender.BORDER_ZERO)
                            },

                    // validParamValues
                    null
                )
        );
    }


    public int getNumSources() {
        return 0;
    }


    private float calculateDivisor(RenderedImage origImage) {
        int dataType = origImage.getSampleModel().getDataType();
        switch (dataType) {
            case DataBuffer.TYPE_BYTE:
                return (float)Byte.MAX_VALUE;

            case DataBuffer.TYPE_SHORT:
                return (float)Short.MAX_VALUE;

            case DataBuffer.TYPE_USHORT:
                return 65535f;

            case DataBuffer.TYPE_INT:
                return (float)Integer.MAX_VALUE;

            default:
                return 1f;
        }
    }


    /**
     * Method needed for implementation of {@link RenderedImageFactory} 
     */
    public RenderedImage create(final ParameterBlock pbIn,
            final RenderingHints rhIn) {

        RenderedOp rop;
        ParameterBlock pb;
/*
        ImageLayout origLayout = new FImageLayout(
                (float)origImage.getMinX(),
                (float)origImage.getMinY(),
                (float)origImage.getWidth(),
                (float)origImage.getHeight());
*/

        RenderedImage origImage = (RenderedImage)pbIn.getObjectParameter(0);
        FImageLayout sourceLayout = new FImageLayout(
                (float)origImage.getMinX(),
                (float)origImage.getMinY(),
                (float)origImage.getWidth(),
                (float)origImage.getHeight());

        float sourceAspectRatio =
                (float)origImage.getWidth()/(float)origImage.getHeight();

        Float oIntensityDivisor = (Float)pbIn.getObjectParameter(1);
        FImageLayout targetLayout = (FImageLayout)pbIn.getObjectParameter(2);
        Boolean oPreserveAspectRatio = (Boolean)pbIn.getObjectParameter(3);
        boolean preserveAspectRatio = (oPreserveAspectRatio != null) ?
                oPreserveAspectRatio.booleanValue() : false;
        Float oSourceSize = (Float)pbIn.getObjectParameter(4);
        BorderExtender bExt = (BorderExtender)pbIn.getObjectParameter(5);


        //---------------------------------------------------------------
        //  Rescale the intensities into the [0.0, 1.0] range
        //---------------------------------------------------------------
        float intensityDivisor = oIntensityDivisor != null ?
                oIntensityDivisor.floatValue() : calculateDivisor(origImage);

        pb = new ParameterBlock();
        pb.addSource(origImage);

        if (intensityDivisor != 1f) {
            FloatSampleModel fsm = new FloatSampleModel(
                    origImage.getWidth(),
                    origImage.getHeight(),
                    origImage.getSampleModel().getNumBands());
            ImageLayout il = new ImageLayout();
            il.setSampleModel(fsm);

            RenderingHints rh =
                    new RenderingHints(JAI.KEY_IMAGE_LAYOUT, il);
            pb.add(new double[] {1.0/(double)intensityDivisor} );
            rop = JAI.create("Rescale", pb, rh);
        }
        else {
            rop = JAI.create("Null", pb);
        }

        //---------------------------------------------------------------
        //  Check to see if the image needs to be truncated
        //  (would only occur if the size is to be preserved).
        //---------------------------------------------------------------
        if (oSourceSize != null) {
            float sourceTargetSize = oSourceSize.floatValue();
            float sourceTargetHeight = sourceTargetSize/(float)Math.sqrt(
                    sourceAspectRatio * sourceAspectRatio + 1);
            float sourceTargetWidth = sourceTargetHeight * sourceAspectRatio;

            float heightTrim = sourceTargetHeight - targetLayout.getHeight();
            float widthTrim = sourceTargetWidth - targetLayout.getWidth();
            
            if (heightTrim > 0f || widthTrim > 0f) {
                if (heightTrim < 0f) {
                    heightTrim = 0f;
                }

                if (widthTrim < 0f) {
                    widthTrim = 0f;
                }

                // Change source info to reflect truncated image
                sourceLayout = new FImageLayout(
                        (float)origImage.getMinX() + 0.5f * widthTrim,
                        (float)origImage.getMinY() + 0.5f * heightTrim,
                        (float)origImage.getWidth() - widthTrim,
                        (float)origImage.getHeight() - heightTrim);

                pb = new ParameterBlock();
                pb.addSource(rop);
                pb.add(sourceLayout.getMinX());
                pb.add(sourceLayout.getMinY());
                pb.add(sourceLayout.getWidth());
                pb.add(sourceLayout.getHeight());
                rop = JAI.create("Crop", pb);
            }
        }

        //---------------------------------------------------------------
        //  Scale the image to the proper size and translate it to the
        //  center of the destination location.
        //---------------------------------------------------------------
        float xScale = targetLayout.getWidth() / sourceLayout.getWidth();
        float yScale = targetLayout.getHeight() / sourceLayout.getHeight();
        float xTrans = targetLayout.getMinX();
        float yTrans = targetLayout.getMinY();
        boolean needsPadding = false;
        if (preserveAspectRatio && xScale != yScale) {
            needsPadding = true;
            xScale = Math.min(xScale, yScale);
            yScale = xScale;
        }

        pb = new ParameterBlock();
        pb.addSource(rop);
        pb.add(xScale);
        pb.add(yScale);
        pb.add(xTrans);
        pb.add(yTrans);
        rop = JAI.create("Scale", pb);

        //---------------------------------------------------------------
        //  Pad remaining area if necessary.
        //---------------------------------------------------------------
        if (needsPadding) {
            float totalHorizPadding = targetLayout.getWidth() -
                    sourceLayout.getWidth() * xScale;
            int leftPadding = (int)(0.5f*totalHorizPadding);
            int rightPadding = (int)totalHorizPadding - leftPadding;

            float totalVertPadding = targetLayout.getHeight() -
                    sourceLayout.getHeight() * yScale;
            int topPadding = (int)(0.5f*totalVertPadding);
            int bottomPadding = (int)totalVertPadding - topPadding;

            pb = new ParameterBlock();
            pb.addSource(rop);
            pb.add(leftPadding);
            pb.add(rightPadding);
            pb.add(topPadding);
            pb.add(bottomPadding);
            pb.add(bExt);
            rop = JAI.create("Border", pb);

            pb = new ParameterBlock();
            pb.addSource(rop);
            pb.add((float)leftPadding);
            pb.add((float)topPadding);
            rop = JAI.create("Translate", pb);
        }

        return rop;
    }

}
