package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;

import javax.media.jai.JAI;
import javax.media.jai.LookupTableJAI;
import javax.media.jai.RenderedOp;

/**
 * This class paints a shape as an outer donut ring of one color
 * and the inside of the ring as another color.
 * But, we'd like the average intensities of the shape to be true to 
 * the specified color.  Below is the derivation of how to calculate the
 * modified intensities.  For convenience of conputation, the portion of
 * the formula that is comprised of values passed into the constructor is
 * combined into constants.  This allows a simple quick calculation as a
 * function of the particular color for each instance of shape.

<hr>

<dl>
<dt><code>I<sub>a</sub></code></dt>
<dd>average intensity that would be used for a uniform fill
</dd>
<dt><code>I<sub>d</sub></code></dt>
<dd>intensity of the donut ring (outer) portion of the feature</dd>
<dt><code>I<sub>h</sub></code></dt>
<dd>intensity of the donut hole (inner) portion of the feature</dd>

<dt><code>A<sub>d</sub></code></dt>
<dd>approximate area of the donut ring (outer) portion of the feature</dd>
<dt><code>A<sub>h</sub></code></dt>
<dd>approximate area of the donut hole (inner) portion of the feature</dd>

<dt><code>R</code></dt>
<dd>ratio of the donut hole (inner) intensity / donut ring (outer) intensity,
the intensityRatio parameter passed into the constructor
</dd>

<dt><code>b<sub>x</sub></dt>
<dd>the x dimension (width) of the shape's bounding rectangle
</dd>
<dt><code>b<sub>x</sub></dt>
<dd>the y dimension (height) of the shape's bounding rectangle
</dd>

<dt><code>r</code></dt>
<dd>estimated radius of the feature, calculated from the shape's bounding 
rectangle</dd>

<dt><code>w</code>
<dd>width of the donut ring as a fraction of estimated radius,
passed into the constructor</dd>
</dl>

<p>
<code>r = (b<sub>x</sub> + b<sub>y</sub>)/4</code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2> (radius estimated from shape's bounding rectangle)</font>
<br>

<code>R = I<sub>h</sub>/I<sub>d</sub></code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2> (by definition of ratio)</font>
<br>
<code>I<sub>h</sub> = R * I<sub>d</sub></code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2> (rearrange) </font>

<br>
<code>I<sub>a</sub> =
(I<sub>d</sub>*A<sub>d</sub> + I<sub>h</sub>*A<sub>h</sub>) /
(A<sub>d</sub> + A<sub>h</sub>)
</code>

<br>
<code>I<sub>a</sub> =
(I<sub>d</sub>*A<sub>d</sub> + R*I<sub>d</sub>*A<sub>h</sub>) /
(A<sub>d</sub> + A<sub>h</sub>)
</code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2> (substitute </font><code>I<sub>h</sub></code><font size=-2>)
</font>

<br>
<code>I<sub>d</sub> =
I<sub>a</sub>*(A<sub>d</sub> + A<sub>h</sub>) /
(A<sub>d</sub> + R*A<sub>h</sub>)
</code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2>(rearrange)</font>

<br>
<code>
A<sub>h</sub> = &Pi;*(r-r*w)<sup>2</sup>
<br>
A<sub>d</sub> = &Pi;*r<sup>2</sup> - A<sub>h</sub>

<br>
K<sub>d</sub> = 
(A<sub>d</sub> + A<sub>h</sub>) / (A<sub>d</sub> + R*A<sub>h</sub>)
</code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2>(define for convenience)</font>

<br>
<code>
K<sub>h</sub> = K<sub>d</sub> * R
</code>
&nbsp; &nbsp; &nbsp; &nbsp;
<font size=-2>(define for convenience)</font>

<br>
<code>
I<sub>d</sub> = K<sub>d</sub> * I<sub>a</sub>

<br>
I<sub>h</sub> = K<sub>h</sub> * I<sub>a</sub>

</code

 */

public class FeatureTextureDonut implements FeatureTexture {

    private final float widthFraction;
    private final LookupTableJAI lookupTable;

    /**
     * Constructor
     *
     * @param widthFraction  Width of the donut ring as a fraction of the
     *          entire radius of the feature.
     * @param intensityRatio  ratio of inside intensity / donut ring intensity
     */
    public FeatureTextureDonut(float widthFraction, float intensityRatio) {
        this.widthFraction = widthFraction;

        // omit the PI terms in calculation since they cancel out anyway
        float donutConstant =
                1f/(1f + (1f-widthFraction) * (1f-widthFraction) * 
                (intensityRatio - 1f));
        float holeConstant = intensityRatio * donutConstant;
        lookupTable = createLookup(holeConstant, donutConstant);
    }


    public RenderedImage createTexture(Shape featureShape, int numChannels) {
        ParameterBlock pb;
        Rectangle2D bounds = featureShape.getBounds2D();

        BufferedImage bi = new BufferedImage((int)bounds.getWidth(),
                (int)bounds.getHeight(), BufferedImage.TYPE_BYTE_BINARY);

        AffineTransform translater =
                AffineTransform.getTranslateInstance(
                        -bounds.getX(), -bounds.getY());
        Shape movedShape = translater.createTransformedShape(featureShape);
        Graphics2D g = bi.createGraphics();
        g.setColor(Color.white);
        g.fill(getHoleShape(movedShape));

        pb = new ParameterBlock();
        pb.addSource(bi);
        pb.add(lookupTable);
        RenderedOp rop = JAI.create("lookup", pb);

        pb = new ParameterBlock();
        pb.addSource(rop);
        pb.add((float)bounds.getMinX());
        pb.add((float)bounds.getMinY());
        rop = JAI.create("translate", pb);

        return rop;
    }


    // return the inner part of the shape, given the original shape.
    private Area getHoleShape(Shape feature) {
        Rectangle bounds = feature.getBounds();
        float radius = (float)(bounds.getWidth() + bounds.getHeight())/4f;
        
        BasicStroke bs = new BasicStroke(2f * widthFraction * radius,
                BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);
        Area workArea = new Area(feature);
        workArea.subtract( new Area(bs.createStrokedShape(feature)) );

        return workArea;
    }


    static private LookupTableJAI createLookup(float holeConstant,
            float donutConstant) {
        float[] values = new float[256];
        //Arrays.fill(values, donutConstant);
        //values[255] = holeConstant;
        values[0] = donutConstant;
        values[1] = holeConstant;
        return new LookupTableJAI(values);
    }

}
