package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.RenderingHints;
import java.awt.geom.Point2D;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.awt.image.renderable.RenderedImageFactory;
import java.util.Arrays;

import javax.media.jai.JAI;
import javax.media.jai.OperationDescriptor;
import javax.media.jai.OperationDescriptorImpl;
import javax.media.jai.OperationRegistry;
import javax.media.jai.ParameterBlockJAI;
import javax.media.jai.ParameterListDescriptor;
import javax.media.jai.ParameterListDescriptorImpl;
import javax.media.jai.RenderedOp;
import javax.media.jai.registry.RIFRegistry;


/**
 *  An operation for creating the aspect of the image including features
 *  specific painting.  This image can then be composited with background
 *  effects and various modifications using JAI operation.s
 */
public class FeaturesImageOpDesc extends OperationDescriptorImpl
        implements RenderedImageFactory {

    public static String OPERATION_NAME = "featuresimage";
    private static boolean isInitialized = false;

    static {
        FeaturesImageOpDesc.init();
    }

    public static void init() {
        if (isInitialized)
            return;

        isInitialized = true;
        OperationRegistry opReg =
                JAI.getDefaultInstance().getOperationRegistry();
        OperationDescriptor od = new FeaturesImageOpDesc();

        opReg.registerDescriptor(od);
        RIFRegistry.register(opReg, od.getName(), OPERATION_NAME,
                             (RenderedImageFactory)od);
    }


   public FeaturesImageOpDesc( ) {

        super(
                // resources
                new String[][] {
                     { "GlobalName",  OPERATION_NAME },
                     { "LocalName",  OPERATION_NAME },
                     { "Vendor",  "Attotron Biosensor Corporation" },
                     { "Description",
                          "The features component image for a zone" },
                     { "DocURL",  "" },
                     { "Version",  "0.9" },
                     { "arg0Desc", "image dimensions" },
                     { "arg1Desc", "number of channels" },
                     { "arg2Desc", "(global) features specifications" },
                     { "arg3Desc", "zones information" }
                },

                // supported modes
                new String[] { "Rendered" },

                // source names
                null,   // use default naming convention "source0", ...

                // source classes
                new Class[][] { { } },

                new ParameterListDescriptorImpl(
                    // descriptor ??
                    new Object(),

                    // paramNames
                    new String[] { "numberOfChannels", "maDimensions", 
                            "featuresSpecs", "zonesInfo" },

                    // paramClasses
                    new Class[]
                         { NumberOfChannels.class, MaDimensions.class, 
                           FeaturesSpecs.class, ZonesInfo.class },

                    // paramDefaults
                    new Object[] {
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT },

                    // validParamValues
                    null
                )
        );
    }


    public int getNumSources() {
        return 0;
    }


    /**
     * Method needed for implementation of {@link RenderedImageFactory}
     */
    public RenderedImage create(final ParameterBlock pbIn,
            final RenderingHints rh)  {

        int numChannels =
                ((NumberOfChannels)pbIn.getObjectParameter(0)).numberOfChannels;
        
        MaDimensions mDimensions = (MaDimensions)pbIn.getObjectParameter(1);
        FeaturesSpecs featuresSpecs = (FeaturesSpecs)pbIn.getObjectParameter(2);
        ZonesInfo zonesInfo = (ZonesInfo)pbIn.getObjectParameter(3);

        // A parameter block for creating a mosaic of the zones.
        ParameterBlockJAI pjb = new ParameterBlockJAI("Mosaic");
        pjb.setParameter("sourceThreshold", new double[][] { { 1e-50 } } );

        int numZones = zonesInfo.getNumZones();
        for (int z=0; z<numZones; ++z) {
            Point2D zoneLocation = zonesInfo.getZoneLocation(z);
            float[][][] featuresIntensities = zonesInfo.getZoneIntensities(z);
            ZoneCharacteristics zoneCharacteristics =
                    zonesInfo.getZoneCharacteristics(z);
            int numRows = featuresIntensities.length;
            int numCols = featuresIntensities[0].length;

            ParameterBlockJAI zpjb = new ParameterBlockJAI("ZoneFeaturesImage");
            zpjb.setParameter("ZoneIndex", z);
            zpjb.setParameter("NumRows", numRows);
            zpjb.setParameter("NumCols", numCols);
            zpjb.setParameter("NumberOfChannels", numChannels);
            zpjb.setParameter("FeaturesSpecs", featuresSpecs);
            zpjb.setParameter("FeaturesIntensities", featuresIntensities);
            zpjb.setParameter("ZoneCharacteristics", zoneCharacteristics);
            RenderedOp zoneImage = JAI.create("ZoneFeaturesImage", zpjb);

            ParameterBlockJAI pbTranslated = new ParameterBlockJAI("Translate");
            pbTranslated.addSource(zoneImage);
            pbTranslated.setParameter("xTrans", (float)zoneLocation.getX());
            pbTranslated.setParameter("yTrans", (float)zoneLocation.getY());
            RenderedOp translatedZone = JAI.create("Translate", pbTranslated);

            pjb.addSource(translatedZone);
        }

        ParameterBlock pb = new ParameterBlock();
        Float[] bgValues = new Float[numChannels];
        Arrays.fill(bgValues, new Float(0f));
        pb.add(mDimensions.width);
        pb.add(mDimensions.height);
        pb.add(bgValues);
        pjb.addSource(JAI.create("constant", pb));

        // RenderingHints rh = null;
        return (rh == null) ? JAI.create("Mosaic", pjb) :
                JAI.create("Mosaic", pjb, rh);
    }

}
