package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.RenderingHints;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.awt.image.renderable.RenderedImageFactory;

import javax.media.jai.BorderExtender;
import javax.media.jai.JAI;
import javax.media.jai.KernelJAI;
import javax.media.jai.OperationDescriptor;
import javax.media.jai.OperationDescriptorImpl;
import javax.media.jai.OperationRegistry;
import javax.media.jai.ParameterListDescriptorImpl;
import javax.media.jai.registry.RIFRegistry;

/**
 *  An operation for Gaussian blur.
 */
public class GBlurOpDesc  extends OperationDescriptorImpl
                           implements RenderedImageFactory  {

    private static boolean isInitialized = false;
    public static final String OPERATION_NAME = "GBlur";

    public static void init() {
        if (isInitialized)
            return;

        isInitialized = true;

        OperationRegistry opReg = 
                JAI.getDefaultInstance().getOperationRegistry();
        OperationDescriptor od = new GBlurOpDesc();
                
        opReg.registerDescriptor(od);
        RIFRegistry.register(opReg, od.getName(), OPERATION_NAME,
                             (RenderedImageFactory)od);
    }


    public GBlurOpDesc() {
        super(
                // resources
                new String[][] {
                     { "GlobalName",  OPERATION_NAME },
                     { "LocalName",  "GBlur" },
                     { "Vendor",  "Attotron Biosensor Corporation" },
                     { "Description",  "Gaussian blur" },
                     { "DocURL",  "" },
                     { "Version",  "0.9" },
                     { "arg0Desc",  "Size of blur (kernel dimension)" },
                     { "arg1Desc",  "Standard deviation of blur" }
                },

                // supported modes
                new String[] { "Rendered" },

                // source names
                null,   // use default naming convention "source0", ...

                // source classes
                new Class[][] { { RenderedImage.class } },

                new ParameterListDescriptorImpl(
                    // descriptor ??
                    new Object(),

                    // paramNames
                    new String[] { "Size", "StdDev" },

                    // paramClasses
                    new Class[] { Integer.class, Float.class },
                
                    // paramDefaults
                    null,

                    // validParamValues
                    null
                )
        );

    }


    protected boolean validateSources(
            final String modeName,
            final ParameterBlock param,
            final StringBuffer message) {

        if (super.validateSources(modeName, param, message))  {
            return true;
        }
        
        return false;
    }


    protected boolean validateParameters(
            final String modeName,
            final ParameterBlock param,
            final StringBuffer message) {
        if (super.validateParameters(modeName, param, message))  {
            return true;
        }
        
        return false;
    }


    //  This is for RenderedImageFactory
    public RenderedImage create(final ParameterBlock params, 
                                final RenderingHints hints) {

        RenderedImage inImage = (RenderedImage)params.getSource(0);
        int kernelSize = params.getIntParameter(0);
        float stdDev = params.getFloatParameter(1);

        RenderingHints rHints = new RenderingHints(
                JAI.KEY_BORDER_EXTENDER,
                BorderExtender.createInstance(BorderExtender.BORDER_COPY)
                );
        KernelJAI kernel = genKernel(kernelSize, stdDev);
        ParameterBlock pb = new ParameterBlock();
        pb.addSource(inImage);
        pb.add(kernel);
        RenderedImage blurOp = JAI.create("convolve", pb, rHints);

        return blurOp;
    }


    static private KernelJAI genKernel(int size, float stdDev) {
        double nominator = 2 * stdDev * stdDev;
        float[] values = new float[size*size];

        int center = (size - 1) / 2;
        int limit = size - 1;
        int xx;
        int yy;
        float sum = 0f;
        float value = 0f;

        for (int y = 0; y < size; y++) {
            for (int x = 0; x < size; x++) {
                if ((y <= center) && (x <= center)) {
                    if (x >= y) {
                        //calculate new value
                        xx = center - x;
                        yy = center - y;
                        value = (float) Math.exp(-(xx*xx + yy*yy) / nominator);
                        values[(y*size)+x] = value;
                        sum += value;
                    } else {
                        //copy existing value
                        value = values[(x*size)+y];
                        values[(y*size)+x] = value;
                        sum += value;
                    }
                } else {
                    xx = x;
                    yy = y;
                    if (yy > center) yy = limit - yy;
                    if (xx > center) xx = limit - xx;
                    value = values[(yy*size)+xx];
                    values[(y*size)+x] = value;
                    sum += value;
                }
            }
        }

        for (int i = 0; i < values.length; i++) {
            values[i] = values[i] / sum;
        }

        return new KernelJAI(size, size, values);
    }

}
