package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.RenderingHints;
import java.awt.image.DataBuffer;
import java.awt.image.PixelInterleavedSampleModel;
import java.awt.image.renderable.ParameterBlock;

import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.RenderedOp;

/**
 *  An image file save type that stores one channel per file, with integers
 *  for pixel values.
 */
public class GrayscaleTiffs extends ScanSaveType {

    static public final String defaultFilenameExtension = "tif";

    private final int numBits;

    public GrayscaleTiffs(CommonScanSaveParams commonParams, int numBits) {
        super(commonParams);
        this.numBits = numBits;
    }


    public GrayscaleTiffs(CommonScanSaveParams commonParams) {
        this(commonParams, 16);
    }


    /**
     * Create grayscale tiff image(s) from the working microarray image,
     * one image file for each channel of the experiment.
     * Before calling this method, the image has already had noise, blur,
     * and gain taken into effect.  The input image has a sample datatype 
     * of float, with saturation = 1.0.
     *
     * @param image     A rendered image containing all the channels.
     * @param fileStem  The stem of the filename; "_1", "_2", ... will be added
     *         for the channels.
     */
    public void createFile(RenderedOp image, String fileStem) {

        ParameterBlock pb;
        int numBands = image.getSampleModel().getNumBands();

        int dataType = (numBits <= 8) ? DataBuffer.TYPE_BYTE :
                       (numBits <= 16) ? DataBuffer.TYPE_USHORT :
                       DataBuffer.TYPE_INT;

        double maxSampleValue = Math.pow(2.0, (double)numBits) - 1.0;

        ImageLayout il = new ImageLayout();
        int[] bandOffsets = new int[numBands];
        for (int b=0; b<numBands; ++b) {
            bandOffsets[b] = b;
        }
        PixelInterleavedSampleModel sampleModel =
                new PixelInterleavedSampleModel(dataType,
                image.getWidth(), image.getHeight(),
                numBands, numBands * image.getWidth(), bandOffsets);
        il.setSampleModel(sampleModel);
        RenderingHints rh = new RenderingHints(JAI.KEY_IMAGE_LAYOUT, il);

        //  Change the data type (dynamic range) to match final destination.
        pb = new ParameterBlock();
        pb.addSource(image);
        pb.add( new double[] {maxSampleValue} );
        image = JAI.create("MultiplyConst", pb, rh);

        //  Truncate values with intensities over the top and
        //  make negative intensities zero.
        pb = new ParameterBlock();
        pb.addSource(image);
        pb.add( new double[] {0.0} );
        pb.add( new double[] {maxSampleValue,maxSampleValue} );
        image = JAI.create("Clamp", pb);

        for (int band=0; band<numBands; ++band) {
            String fName = imagePath(fileStem, 
                    (numBands == 1) ? "" : "_" + String.valueOf(band+1) );
            createTiff(image, band, fName);
        }
    }


    /**
     * Create an tiff file for a given band (channel).
     *
     * @param inImage  The input working image from which to extract the band.
     * @param band  The number (index) of the band to extract.
     * @param fileName  The complete file name of where to save the TIFF.
     */
    private void createTiff(RenderedOp inImage, int band, String fileName) {
        ParameterBlock pb;

        pb = new ParameterBlock();
        pb.addSource(inImage);
        pb.add( new int[]{band} );
        inImage = JAI.create("BandSelect", pb);

        pb = new ParameterBlock();
        pb.addSource(inImage);
        pb.add(fileName);
        pb.add("tiff");
        JAI.create("FileStore", pb);
    }

}
