package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.Point;
import java.awt.geom.Point2D;

/**
 * This class encapsulates the layout of a microarray with zones on a uniform
 * grid, and where each zone has the same number of rows and columns.
 */
public class GridMaLayout  implements MaLayout {

    private final int numMetacols;
    private final int numMetarows;
    private final int numRows;
    private final int numCols;
    private final float leftMargin;
    private final float rightMargin;
    private final float topMargin;
    private final float bottomMargin;
    private final float gapBetweenZonesX;
    private final float gapBetweenZonesY;
    private final float featureCellSizeX;
    private final float featureCellSizeY;

    private float zoneWidth;
    private float zoneHeight;
    private Point[] zoneGridLocations = null;


    public GridMaLayout(int numMetacols, int numMetarows,
            int numCols, int numRows,
            float leftMargin, float rightMargin,
            float topMargin, float bottomMargin,
            float gapBetweenZonesX, float gapBetweenZonesY,
            float featureCellSizeX, float featureCellSizeY ) {
        this.numMetacols = numMetacols;
        this.numMetarows = numMetarows;
        this.numRows = numRows;
        this.numCols = numCols;
        this.leftMargin = leftMargin;
        this.rightMargin = rightMargin;
        this.topMargin = topMargin;
        this.bottomMargin = bottomMargin;
        this.gapBetweenZonesX = gapBetweenZonesX;
        this.gapBetweenZonesY = gapBetweenZonesY;
        this.featureCellSizeX = featureCellSizeX;
        this.featureCellSizeY = featureCellSizeY;
        
        this.zoneWidth = featureCellSizeX * numCols;
        this.zoneHeight = featureCellSizeY * numRows;
    }


    public Object getScaled(float pixelSize) {
        GridMaLayout gml = new GridMaLayout(numMetacols, numMetarows,
            numCols, numRows,
            leftMargin/pixelSize, rightMargin/pixelSize,
            topMargin/pixelSize, bottomMargin/pixelSize,
            gapBetweenZonesX/pixelSize, gapBetweenZonesY/pixelSize,
            featureCellSizeX/pixelSize, featureCellSizeY/pixelSize);
        if (zoneGridLocations != null) {
            gml.addZonesGridLocations(zoneGridLocations);
        }
        return gml;
    }


    public void addZonesGridLocations(Point[] zoneGridLocations) {
        this.zoneGridLocations = zoneGridLocations;
    }


    public MaDimensions getDimensions() {
        return new MaDimensions(getWidth(), getHeight());
    }


    public float getWidth() {
        return leftMargin + rightMargin +
                (float)numMetacols * zoneWidth +
                (float)(numMetacols-1) * gapBetweenZonesX;
    }


    public float getHeight() {
        return topMargin + bottomMargin +
                (float)numMetarows * zoneHeight +
                (float)(numMetarows-1) * gapBetweenZonesY;
    }


    public Point2D getZoneLocation(int zoneIndex) {
        int metarow;
        int metacol;

        if (zoneGridLocations != null) {
            metarow = zoneGridLocations[zoneIndex].y;
            metacol = zoneGridLocations[zoneIndex].x;
        }
        else {
            metarow = zoneIndex/numMetacols;
            metacol = zoneIndex%numMetacols;
        }

        return new Point2D.Float(
                leftMargin + (float)metacol * (zoneWidth + gapBetweenZonesX),
                topMargin + (float)metarow * (zoneHeight + gapBetweenZonesY));
    }

}
