package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.io.IOException;
import java.io.LineNumberReader;
import java.io.Reader;
import java.util.Vector;
import java.util.regex.Matcher;
import java.util.regex.Pattern;


/**
 *  A flexible class to read feature intensities data from a flat file.
 *  
 *  Input data includes:  zone index and/or metarow/metacol, row, column,
 *  and intensities values.
 *
 */
public class IntensitiesFileReader implements IntensitiesReader {

    // IDs for specifying existence and order of fields in line of input data.
    static public final int ID_Ignore = 0;
    static public final int ID_ZoneIndex = 1;
    static public final int ID_Metarow = 2;
    static public final int ID_Metacol = 3;
    static public final int ID_Row = 4;
    static public final int ID_Col = 5;

    static private final Pattern tabPattern = Pattern.compile("\t");
    
    private RawIntensitiesRecord[] data = null;

    private final LineNumberReader bufRdr;
    private final int[] columnIDs;
    private boolean onlyOneZone;
    private boolean includesZoneIndex;
    private boolean includesMetaCoords;
    private Pattern pattern;
    private Pattern commentLinePattern;
    private int numMetacols = 0;

    private int numChannels = 0;   // set after reading first record



    public IntensitiesFileReader(Reader rdr) throws DatMissingColumn {
        this(rdr, null);
    }


    public IntensitiesFileReader(Reader rdr, int[] columnIDs)
            throws DatMissingColumn {
        this(rdr, columnIDs, null, false);
    }


    public IntensitiesFileReader(Reader rdr, int[] columnIDs, Pattern pattern, 
            boolean onlyOneZone) throws DatMissingColumn {
        this.onlyOneZone = onlyOneZone;
        if (columnIDs == null) {
            columnIDs = onlyOneZone ? new int[] {ID_Row, ID_Col} :
                    new int[] {ID_ZoneIndex, ID_Row, ID_Col};
        }
        bufRdr = new LineNumberReader(rdr);
        examineColumns(columnIDs);
        this.columnIDs = columnIDs;
        this.pattern = pattern;

        commentLinePattern = Pattern.compile("\\A#");
    }


    public boolean includesMetaCoords() {
        return includesMetaCoords;
    }


    public boolean includesZoneIndex() {
        return includesZoneIndex;
    }


    /**
     *  This method examines the specification of the columns (or captures)
     *  in the input data.  If neither the zone index nor the metarow/metacol
     *  is included in the input, an exception is thrown.
     *
     *  @param columnIDs  the list of ids of fields specified for columns in
     *      the input data.
     */
    private void examineColumns(int[] columnIDs) throws DatMissingColumn {
        boolean includesMetarow = false;
        boolean includesMetacol = false;

        includesZoneIndex = false;
        for (int i=0; i<columnIDs.length; ++i) {
            switch (columnIDs[i]) {
                case ID_ZoneIndex:
                    includesZoneIndex = true;
                    break;

                case ID_Metarow:
                    includesMetarow = true;
                    break;

                case ID_Metacol:
                    includesMetacol = true;
                    break;

                case ID_Ignore:
                    break;
            }
        }

        includesMetaCoords = (includesMetarow && includesMetacol);
        if (!onlyOneZone && !includesZoneIndex && !includesMetaCoords) {
            throw new DatMissingColumn();
        }
    }


    public void setCommentLinePattern(Pattern commentLinePattern) {
        this.commentLinePattern = commentLinePattern;
    }


    public RawIntensitiesRecord[] readData() throws DatException {
        String lineStr;
        Vector vData = new Vector();
        try {
            while ((lineStr = bufRdr.readLine()) != null) {
                if (commentLinePattern == null ||
                        !commentLinePattern.matcher(lineStr).find()) {
                    vData.add(processLine(lineStr));
                }
            }
        }
        catch (IOException e) {
            throw new DatException("Error in reading raw intensities data", e);
        }

        return (RawIntensitiesRecord[]) vData.toArray(new RawIntensitiesRecord[
                vData.size()]);
    }


    public RawIntensitiesRecord[] getData() throws DatException {
        if (data == null) {
            data = readData();
        }

        return data;
    }


    private RawIntensitiesRecord processLine(String str)
            throws DatException  {
        String[] tokens = null;
        if (pattern != null) {
            Matcher matcher = pattern.matcher(str);
            int numGrps = matcher.groupCount();
            tokens = new String[numGrps];
            for (int i=0; i<numGrps; ++i) {
                tokens[i] = matcher.group(i);
            }
        }
        else {
            tokens = tabPattern.split(str);
        }

        int numChannelsInData = tokens.length - columnIDs.length;
        if (numChannelsInData < 1) {
            // throw
        }
        else if (numChannels == 0) {    // first record, numChannels not yet set
            numChannels = numChannelsInData;
        }
        else {
            // throw
        }

        RawIntensitiesRecord rd = new RawIntensitiesRecord();
        if (onlyOneZone) {
            rd.zoneIndex = 0;
        }

        int col;
        for (col=0; col<columnIDs.length; ++col) {
            int intVal = -1;
            try {
                intVal = Integer.parseInt(tokens[col]);
            }
            catch (NumberFormatException e) {
                throw new DatException("Raw intensities data: " +
                        "Integer format error", e);
            }
            if (intVal < 0) {
                throw new DatException("Raw intensites data: " +
                        "integer must be positive");
            }

            switch (columnIDs[col]) {
                case ID_ZoneIndex:
                    rd.zoneIndex = intVal;
                    break;

                case ID_Metarow:
                    rd.metarow = intVal;
                    break;

                case ID_Metacol:
                    rd.metacol = intVal;
                    break;

                case ID_Row:
                    rd.row = intVal;
                    break;

                case ID_Col:
                    rd.col = intVal;
                    break;
            }
        }

        int colFirstIntensity = col;
        rd.intensities = new float[tokens.length-colFirstIntensity];

        for (; col<tokens.length; ++col) {
            try {
                rd.intensities[col-colFirstIntensity] =
                    Float.parseFloat(tokens[col]);
            }
            catch (NumberFormatException e) {
                throw new DatException("Raw intensities data float format " + 
                        " error", e);
            }
            if (rd.intensities[col-colFirstIntensity] < 0f) {
                throw new DatException("Raw intensities must not be negative");
            }
        }

        return rd;
    }

}
