package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.geom.Point2D;
import java.awt.image.RenderedImage;
import java.util.Iterator;
import java.util.Vector;

import javax.media.jai.ImageLayout;
import javax.media.jai.ParameterBlockJAI;
import javax.media.jai.RenderedOp;

/**
 *  A version of {@link ParameterBlockJAI} that handles recursive
 *  scaling to pending pixel size of parameter and source images.
 */
public class PendingParameterBlockJAI extends ParameterBlockJAI {

    private String pixelSizeParam;
    private Vector divideToScaleList = null;
    private Vector paramScalers = null;


    public PendingParameterBlockJAI(String operatorName ) {
        super(operatorName);
    }


    /**
     *  This method applies the pixel size recursively to all parameters
     *  and image sources.
     *  @param pixelSize  the pixel size in microns (defines resolution)
     */
    public ParameterBlockJAI manifestResolution(float pixelSize) {
        ParameterBlockJAI pbj = (ParameterBlockJAI)super.clone();
        Iterator iter;
        if (pixelSizeParam != null) {
            pbj.setParameter(pixelSizeParam, pixelSize);
        }

        if (paramScalers != null) {
            iter = paramScalers.iterator();
            while (iter.hasNext()) {
                PScaler pScaler = (PScaler)iter.next();
                pbj.setParameter(pScaler.paramName,
                        pScaler.scaler.getScaledParameter(pixelSize,
                        getObjectParameter(pScaler.paramName)) );
            }
        }

        if (divideToScaleList != null) {
            iter = divideToScaleList.iterator();
            while (iter.hasNext()) {
                String paramName = (String)iter.next();
                pbj.setParameter(paramName, divideToScale(pixelSize,
                        getObjectParameter(paramName)) );
            }
        }

        Vector params = getParameters();
        Object[] aParams = new Object[params.size()];
        aParams = params.toArray(aParams);
        for (int i=0; i<aParams.length; ++i) {
            if (aParams[i] instanceof SelfScaler) {
                pbj.set(((SelfScaler)aParams[i]).getScaled(pixelSize), i);
            }
        }
   

        Vector sources = getSources();
        Object[] aSources = new Object[sources.size()];
        
        aSources = sources.toArray(aSources);
        for (int i=0; i<aSources.length; ++i) {
            if (aSources[i] instanceof PendingRenderedOp) {
                PendingRenderedOp ropp = (PendingRenderedOp)aSources[i];
                RenderedOp rop = ropp.manifestResolution(pixelSize);
                pbj.setSource(rop, i);
            }
        }

        return pbj;
    }


    /**
     * This sets the specified parameter (by name) to be set to the
     * pixel size once its known.
     *
     * @param paramName the name of the parameter
     */
    public void setPixelSizeParamName(String paramName) {
        pixelSizeParam = paramName;
    }


    /**
     *  A call to this method sets it up so that the given named parameter
     *  is divided by the pixel size when it becomes available.
     */
    public void setParameterDivideToScale(String paramName) {
        if (divideToScaleList == null) {
            divideToScaleList = new Vector();
        }
        divideToScaleList.add(paramName);
    }


    /**
     *  A utility method for dividing a parameter by the pixelSize. 
     *
     * @param pixelSize  the size of pixels in microns (defines resolution)
     * @param unscaled  the input parameter before scaling
     *
     * @return  the scaled parameter
     */
    private static Object divideToScale(float pixelSize, Object unscaled) {
       if (unscaled instanceof Float) {
            Float fVal = (Float)unscaled;
            return new Float( fVal.floatValue() / pixelSize );
        }

        else if (unscaled instanceof Integer) {
            Integer iVal = (Integer)unscaled;
            return new Integer( (int)((float)iVal.intValue() / pixelSize) );
        }

        else if (unscaled instanceof Point2D) {
            Point2D pVal = (Point2D)unscaled;
            double x = pVal.getX();
            double y = pVal.getY();
            return new Point2D.Double(x/pixelSize, y/pixelSize);
        }

        return null;
    }


    /**
     *  Get the {@link javax.media.jai.ImageLayout} object that describes
     *  the source image.
     *
     * @param sourceNum  the index of the image source
     * @return the {@link ImageLayout} for the specified source
     */
    public ImageLayout getSourceImageLayout(int sourceNum) {
        if (sourceNum >= getNumSources()) {
            throw new RuntimeException(
                    "Trying to propagate a non-existent source");
        }
        RenderedImage image = (RenderedImage)getSource(sourceNum);
        return new ImageLayout(image.getMinX(), image.getMinY(),
                image.getWidth(), image.getHeight());

    }


    /**
     *  Assign a {@link JAIParamScaler} to scale a particular parameter 
     *  when the pixel size is known.
     */
    public void setParameterScaler(String paramName, JAIParamScaler scaler) {
        if (paramScalers == null) {
            paramScalers = new Vector();
        }
        PScaler ps = new PScaler();
        ps.paramName = paramName;
        ps.scaler = scaler;
        paramScalers.add(ps);
    }


    class PScaler {
        String paramName;
        JAIParamScaler scaler;
    }
}
