package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.RenderingHints;
import java.awt.image.RenderedImage;

import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.ParameterBlockJAI;
import javax.media.jai.RenderedOp;

/**
 * This class is analogous to {@link javax.media.jai.RenderedOp},
 * but in addition it handles recursive application of pixel size.
 * Once the pixel size is known, the {@link #manifestResolution} method 
 * is called to create a {@link javax.media.jai.RenderedOp} with the
 * appropriate pixel size.
 */
public class PendingRenderedOp implements SelfScaler {

    private PendingParameterBlockJAI pbjp;
    private RenderingHints rhp;

    private FImageLayout imageLayout = null;
    private boolean propagateImageLayout = false;
    private int propagateSourceNum = 0;

    private RenderedOp manifestImage = null;
    private float manifestPixelSize = 0f;
    private float pendingPixelSize = 0f;
    private int numChannels = 0;


    /**
     *  Returns a scaled version of this object.
     */
    public Object getScaled(float pixelSize) {
        pendingPixelSize = pixelSize;
        return this;
    }


    /**
     *
     * @return the image layout of the image
     */
    public FImageLayout getLayout() {
        if (pendingPixelSize == 0f) {
            throw new RuntimeException("Shouldn't get here");
        }
        
        RenderedOp rop = manifestResolution(pendingPixelSize);
        
        return new FImageLayout(rop);
    }


    public PendingRenderedOp(PendingParameterBlockJAI pbjp) {
        this.pbjp = pbjp;
    }


    public PendingRenderedOp(PendingParameterBlockJAI pbjp, RenderingHints rh) {
        this.pbjp = pbjp;
        this.rhp = rh;
        if (rh.containsKey(JAI.KEY_IMAGE_LAYOUT)) {
            imageLayout =
                    new FImageLayout((ImageLayout)rh.get(JAI.KEY_IMAGE_LAYOUT));
        }
    }


    /**
     * Configure the number of channels for the output.  This influences the
     * {@link java.awt.RenderingHints} passed to JAI.create.
     *
     * @param numChannels  the number of channels desired for the output
     * @return the image layout of the image
     */
    public void setNumberOfChannels(int numChannels) {
        this.numChannels = numChannels;
    }


    public void setImageLayout(FImageLayout imageLayout) {
        this.imageLayout = imageLayout;
    }


    public void propagateImageLayout(int sourceNum) {
        propagateImageLayout = true;
        propagateSourceNum = sourceNum;
    }


    private ImageLayout getPropagatedImageLayout(ParameterBlockJAI pbj,
            int srcNum, float pixelSize) {
        if (srcNum >= pbj.getNumSources()) {
            throw new RuntimeException(
                    "Trying to propagate a non-existent source");
        }
        Object obj = pbj.getSource(srcNum);

        int minX, minY, width, height;

        if (obj instanceof PendingRenderedOp) {
            obj = ((PendingRenderedOp)obj).manifestResolution(pixelSize);
        }

        RenderedImage image = (RenderedImage)obj;
        return new ImageLayout(image.getMinX(), image.getMinY(),
                image.getWidth(), image.getHeight());
    }


    private RenderingHints getRenderingHints(float pixelSize) {
    
        ImageLayout scaledImageLayout = null;
        if (imageLayout != null) {
            FImageLayout scil =
                    ((FImageLayout)imageLayout.getScaled(pixelSize));
            scaledImageLayout = scil.getImageLayout();
        }
        else if (propagateImageLayout) {
            scaledImageLayout = getPropagatedImageLayout(pbjp,
                    propagateSourceNum, pixelSize);
        }

        if (scaledImageLayout != null && numChannels > 0) {
            scaledImageLayout.setSampleModel( new FloatSampleModel(
                    scaledImageLayout.getWidth(null),
                    scaledImageLayout.getHeight(null),
                    numChannels));
        }

        RenderingHints rHints = rhp;
        if (scaledImageLayout != null) {

            if (rHints == null) {
                rHints = new RenderingHints(JAI.KEY_IMAGE_LAYOUT,
                        scaledImageLayout);
            }
            else {
                rHints.put(JAI.KEY_IMAGE_LAYOUT, scaledImageLayout);
            }
        }

        return rHints;
    }


    /**
     * This method invokes the application of pixel size.
     *
     * @param pixelSize  the desired pixel size in microns
     * @return  the image with the requested pixel size manifested.
     */
    public RenderedOp manifestResolution(float pixelSize) {

        if (manifestImage == null || manifestPixelSize != pixelSize) {
            ParameterBlockJAI pbj = pbjp.manifestResolution(pixelSize);
    
            RenderingHints rHints = getRenderingHints(pixelSize);

            String operatorName = pbjp.getOperationDescriptor().getName();
    
            manifestPixelSize = pixelSize;
            manifestImage = (rHints == null) ?
                    JAI.create(operatorName, pbj) :
                    JAI.create(operatorName, pbj, rHints);
        }

        return manifestImage;
    }

}
