package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.image.RenderedImage;
import java.awt.image.WritableRaster;
import java.awt.image.renderable.ParameterBlock;
import java.awt.image.renderable.RenderedImageFactory;
import java.util.Arrays;

import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.OperationDescriptor;
import javax.media.jai.OperationDescriptorImpl;
import javax.media.jai.OperationRegistry;
import javax.media.jai.ParameterListDescriptor;
import javax.media.jai.ParameterListDescriptorImpl;
import javax.media.jai.PlanarImage;
import javax.media.jai.PointOpImage;
import javax.media.jai.iterator.RectIter;
import javax.media.jai.iterator.RectIterFactory;
import javax.media.jai.iterator.WritableRectIter;
import javax.media.jai.registry.RIFRegistry;

/**
 *  A JAI operation to modify pixel values, based upon a
 *  {@link PointModifyFunction} passed in.
 */
public class PointModifyFunctionOpDesc extends OperationDescriptorImpl 
        implements RenderedImageFactory {

    static public String OPERATION_NAME = "ModifyFunction";

    private static boolean isInitialized = false;

    public static void init() {
        if (isInitialized)
            return;

        isInitialized = true;
        OperationRegistry opReg =
                JAI.getDefaultInstance().getOperationRegistry();
        OperationDescriptor od = new PointModifyFunctionOpDesc();

        opReg.registerDescriptor(od);
        RIFRegistry.register(opReg, od.getName(), OPERATION_NAME,
                             (RenderedImageFactory)od);
    }


    public PointModifyFunctionOpDesc( ) {

        super(
                // resources
                new String[][] {
                     { "GlobalName",  OPERATION_NAME },
                     { "LocalName",  OPERATION_NAME },
                     { "Vendor",  "Attotron Biosensor Corporation" },
                     { "Description",
                       "Modify intensities, one pixel at a time" },
                     { "DocURL",  "" },
                     { "Version",  "0.9" },
                     { "arg0Desc", "Point Modify Function" },
                     { "arg1Desc", "skip black?" }
                },

                // supported modes
                new String[] { "Rendered" },

                // source names
                null,   // use default naming convention "source0", ...

                // source classes
                new Class[][] { { RenderedImage.class } },

                new ParameterListDescriptorImpl(
                    // descriptor ??
                    new Object(),

                    // paramNames
                    new String[] {
                            "ModifyFunction",
                            "SkipBlack" },

                    // paramClasses
                    new Class[] {
                            PointModifyFunction.class,
                            Boolean.class },

                    // paramDefaults
                    new Object[] {
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            Boolean.FALSE
                            },

                    // validParamValues
                    null
                )
        );
    }


    public int getNumSources() {
        return 1;
    }


    //  This is for RenderedImageFactory
    public RenderedImage create(final ParameterBlock params,
                                final RenderingHints hints) {

        RenderedImage inImage = (RenderedImage)params.getSource(0);
        PointModifyFunction function = (PointModifyFunction)
                params.getObjectParameter(0);
        Boolean bSkipBlack = (Boolean)params.getObjectParameter(1);
        boolean skipBlack = bSkipBlack.booleanValue();

        PointModifyFunctionOp op = new PointModifyFunctionOp(inImage, function);
        if (skipBlack) {
            op.setSkipBlack(skipBlack);
        }

        return op;
    }


    /**
     *  Custom JAI operation that adds Gaussian noise.
     */
    public class PointModifyFunctionOp  extends PointOpImage {
    
        private PointModifyFunction function;
        private boolean skipBlack = false;
    
        /**
         * Constructor, called from RIF
         */
        public PointModifyFunctionOp(RenderedImage source,
                PointModifyFunction function) {
    
            super(source, new ImageLayout(source), null, false);
            this.function = function;
            permitInPlaceOperation();
        }
    
    
        public void setSkipBlack(boolean skipBlack) {
            this.skipBlack = skipBlack;
        }
    
    
        /** Method required for RenderedOp */
        public void computeRect(
                    final PlanarImage[] images,
                    final WritableRaster dest,
                    final Rectangle destRect
                    ) {
    
            final int numBands = images[0].getNumBands();
            float[] samples = new float[numBands];
            float[] zeroes = null;
    
            final RectIter srcIter = RectIterFactory.create(images[0],
                    mapSourceRect(destRect, 0));
    
            final WritableRectIter destIter =
                    RectIterFactory.createWritable(dest, destRect);
    
            if (skipBlack) {
                zeroes = new float[numBands];
                Arrays.fill(zeroes, 0f);
            }
    
            destIter.startLines();
            srcIter.startLines();
            do {
                destIter.startPixels();
                srcIter.startPixels();
                do {
                    samples = srcIter.getPixel(samples);
                    if (!(skipBlack && Arrays.equals(samples, zeroes))) {
                        function.modifyValues(samples);
                    }
                    destIter.setPixel(samples);
                    srcIter.nextPixel();
                } while (!destIter.nextPixelDone());
                srcIter.nextLine();
            } while (!destIter.nextLineDone());
        }
    
    }

}
