package org.cybertory.MSS;

/*-------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
/*
 * PolygonFactory.java
 *
 * Created on December 17, 2003, 11:28 AM
 */

import java.awt.Polygon;
import java.util.Random;
import java.util.Vector;

/**
An object instance of PolygonFactory is a factory to create polygons of a specified
degree of variability.  The method {@link #createPolygon(float)}
produces a {@link java.awt.Polygon} whose approximate radius is specified in
the call, centered at (0,0).

Pseudo-random polygons serve as polygons.

The formulae for calculating polar coordinates
(r<sub>n</sub>, &Theta;<sub>n</sub>) representing the location of 
Vertex<sub>n</sub> with the polygon centered at (0,0) are shown below.
Reference to variables passed into the constructor or method createPolygon
are in boldface.

<hr>
<p>(RADIUS)
<p><code>
r<sub>n</sub> = (1+d<sub>n</sub>) * R
<br>
d<sub>0</sub> = v / (1+s)
<br>
d<sub>n</sub> = v + d<sub>n-1</sub>*s / (1+s)
<br>
</code>

<p>where:
<dl>
<dt><code>R</code></dt>
<dd>is <strong>refRadius</strong> (the reference radius), passed into the 
{@link #createPolygon(float)} method
</dd>
<dt><code>d<sub>n</sub></code></dt>
<dd>is the fractional deviation from the reference radius for vertex n
</dd>
<dt><code>v</code></dt>
<dd>is a random number from a gaussian distribution whose standard deviation 
is specified by <strong>radiusVariability</strong>,
but with a floor (minimum value) of -0.8.  A different value is generated for
the calculation of each vertex.
</dd>
<dt><code>d<sub>n-1</sub></code></dt>
<dd>is the fractional deviation of the previously calculated vertex 
</dd>
<dt><code>s</code></dt>
<dd><strong>radiusStickiness</strong>, how much to weight the previous vertex
radius in the calculation of the current radius; this number must be &gt;= 0
and is usually much less than 1.
</dd>
</dl>
<hr>

<p>(ANGLE)
<p><code>
&Theta;<sub>0</sub> = 0
<br>
&Theta;<sub>n</sub> = &Theta;<sub>n-1</sub> + (1 + v) * 2*&Pi;/n
</code>

<p>where:
<dl>
<dt><code>&Theta;<sub>n</sub></code></dt>
<dd>is the angle of vertex n in radians</dd>
<dt><code>&Theta;<sub>n-1</sub></code></dt>
<dd>is the angle of the previous vertex</dd>
<dt><code>v</code></dt>
<dd>is a random number from a gaussian distribution whose standard deviation 
is specified by <strong>angleVariability</strong>, but with a floor
(minimum value) of -0.9.  A different value is generated for the calculation
of each vertex.
<dt><code>n</code></dt>
<dd>is the <strong>approxNumVertices</strong> (approximate number of vertices) 
</dd>
</dl>

*/

public class PolygonFactory {

    private static float[] radVar =
            new float[] { 0.06f,0.1f,0.2f,1.5f,4f,7f,8f};
    private static float[] radStick =
            new float[] { 0.6f,0.4f,0.4f,0.2f,15f,1f,-0.4f};
    private static int[]   numVert =
            new int[]   { 19,14,10,8,7,6,7};
    private static float[] angVar =
            new float[] { 0.1f,0.2f,0.3f,0.4f,0.6f,0.8f,0.5f};


    private Random rand;
    private final float radiusVariability;
    private final float radiusStickiness;
    private final int   approxNumVertices;
    private final float angleVariability;

    private Vector angleBiases = new Vector();
    
    private float defaultRadius;
    

    /**
     *  Constructor: See class documentation for formulas to see how
     * these parameters are used.
     *
     * @param rand  a random number generator object
     * @param radiusVariability a measure of polygon radius variability
     * @param radiusStickiness how much the radius of one vertex effects the
     *          radius of the subsequent vertex
     * @param approxNumVertices  the approximate number of vertices in the
     *          polygon polygon (affects smoothness)
     * @param angleVariability  a measure of variability of angles between
     *          vertices
     */
    public PolygonFactory(
            Random rand,
            float radiusVariability,
            float radiusStickiness,
            int   approxNumVertices,
            float angleVariability) {

        this.rand = rand;
        this.radiusVariability = radiusVariability;
        this.radiusStickiness = radiusStickiness;
        this.approxNumVertices = approxNumVertices;
        this.angleVariability = angleVariability;
    }


    /**
     *  Constructor with pre-packaged variability parameters.
     *
     * @param variabilityLevel one of the following constants:
     */
    public PolygonFactory(Random rand, int variabilityLevel) {
        this(rand,
                radVar[variabilityLevel-1],
                radStick[variabilityLevel-1],
                numVert[variabilityLevel-1],
                angVar[variabilityLevel-1] );
    }


    public void addAngleBias(AngleBias bias) {
        angleBiases.add(bias);
    }


    private float getRadiusMultiple(float vertexAngle) {
        if (angleBiases.size() == 0) {
            return 1f;
        }

        AngleBias[] biases = new AngleBias[angleBiases.size()];
        angleBiases.toArray(biases);
        return AngleBias.getMultiple(biases, vertexAngle);
    }


    /**
     * Set up a default radius for use with calls to the
     * {@link #createPolygon()} without a radius parameter.
     *
     * @param defaultRadius default radius in pixels
     */
    public void setDefaultRadius(float defaultRadius) {
        this.defaultRadius = defaultRadius;
    }


    public PolarCoords[] createPolygonPolar() {
        return createPolygonPolar(defaultRadius);
    }
    

    public PolarCoords[] createPolygonPolar(float refRadius) {
        float minDeltaRad = -0.9f;
        float minRandAngle = -0.9f;

        float deltaRad;
        float prevDeltaRad = 0f;
        float angle = 0f;
        float radius;
        float randAngle;
        boolean firstVertex = true;

        Vector pCoords = new Vector();

        while (true) {
            deltaRad = ( radiusVariability * (float)rand.nextGaussian()
                        + prevDeltaRad * radiusStickiness)
                        / (1f + radiusStickiness);

            if (deltaRad < minDeltaRad) {
                deltaRad = minDeltaRad;
            }
            prevDeltaRad = deltaRad;
            radius = deltaRad + refRadius;

            if (!firstVertex) {
                randAngle = angleVariability * (float)rand.nextGaussian();
                if (randAngle < minRandAngle) {
                    randAngle = minRandAngle;
                }
                angle += (1f + randAngle) * 
                        ((float)(2*Math.PI)/(float)approxNumVertices);

                if (angle >= 2.0f * (float)Math.PI) {
                    break;
                }
            }

            radius *= getRadiusMultiple(angle);

            pCoords.add( new PolarCoords(angle, radius) );
            firstVertex = false;
        }

        PolarCoords[] retArray = new PolarCoords[pCoords.size()];
        pCoords.toArray(retArray);
        
        return retArray;
    }

    
    /**
     * Creates a polygon with the default radius
     * (which must have been set up with a previous call to 
     * {@link #setDefaultRadius(float)}.
     *
     * @return a {@link java.awt.Polygon} object for the polygon.
     */
    public Polygon createPolygon() {
        return createPolygon(defaultRadius);
    }


    /**
     * Creates a polygon with the specified (approximate) radius.
     *
     * @param refRadius the reference (approximate) radius of the polygon
     * @return a {@link java.awt.Polygon} object for the polygon.
     */
    public Polygon createPolygon(float refRadius)  {

        PolarCoords[] pCoords = createPolygonPolar(refRadius);
        Polygon poly = new Polygon();

        for (int i=0; i<pCoords.length; ++i) {
            poly.addPoint((int)(Math.cos(pCoords[i].angle) * pCoords[i].radius),
                          (int)(Math.sin(pCoords[i].angle) * pCoords[i].radius)
                    );
        }

        return poly;
    }
}
