package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.util.Random;

/**
 *  Custom JAI operation that adds noise as the sum of a series of
 *  specified terms of a polynomial formula, each being a normal distribution.
 *  Each term is specified as an instance of {@link PolynomialNoiseTerm}:

<p><code>
noise = S<sup>exp</sup> * stddev * Z()
</code>

<p>where:
<dl>
<dt><code>S</code></dt>
<dd>is the (signal) intensity</dd>
<dt><code>exp</code></dt>
<dd>is the exponent specified in the PolynomialNoiseTerm</dd>
<dt><code>stddev</code></dt>
<dd>is the standard deviation specified in the PolynomialNoiseTerm</dd>
<dt><code>Z()</code></dt>
<dd>is a normal distribution centered around zero with a stddev of one</dd>
</dl>


 *  Each component contributes a distribution with a mean of zero, so the sum
 *  also has a mean of zero.
 *  
 */
public class PolynomialNoiseFunction  extends PointModifyFunction {

    private Random rand;
    private PolynomialNoiseTerm[] terms;
    private boolean channelsIndependent;


    public PolynomialNoiseFunction(
            Random rand,
            PolynomialNoiseTerm[] terms) {
        this(rand, terms, true);
    }


    /**
     * Constructor, called from RIF
     */
    public PolynomialNoiseFunction(
            Random rand,
            PolynomialNoiseTerm[] terms,
            boolean channelsIndependent) {

        this.rand = (rand == null) ? new Random() : rand;
        this.terms = terms;
        this.channelsIndependent = channelsIndependent;
    }


    public void modifyValues(float[] values) {
        float delta = 0f;
        if (channelsIndependent) {
            float sum = 0f;
            for (int i=0; i<values.length; ++i) {
                sum += values[i];
            }
            delta = getDelta(sum/(float)values.length);
        }
        for (int i=0; i<values.length; ++i) {
            if (!channelsIndependent) {
                delta = getDelta(values[i]);
            }
            values[i] += delta;
        }
    }


    private float getDelta(float srcValue) {
        float sum = 0f;

        for (int i=0; i<terms.length; ++i) {
            sum += terms[i].calc(srcValue, rand);
        }
        return sum;
    }

}
