package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.io.File;
import java.util.HashMap;

import javax.media.jai.JAI;

/**
 *  A library of image files, either in a directory or as specified in a
 *  hash containing a map to the image file locations.
 */
public class RImageFileLibrary implements RImageLibrary {

    private final HashMap fileMap;

    public RImageFileLibrary(HashMap fileMap) {
        this.fileMap = fileMap;
    }


    /**
     *  Initialize a library of all the files in the given directory.
     *  The keys are the unqualified file names.
     *
     * @param directory  the directory containing image files for the library
     */
    public RImageFileLibrary(File directory) {
        this(buildMap(directory, null));
    }


    /**
     *  Initialize a library of the files matching the given pattern string 
     *  in the given directory.
     *  The keys are the unqualified file names.
     *
     * @param directory  the directory containing image files for the library
     * @param patternStr the filename pattern that library members must match
     */
    public RImageFileLibrary(File directory, String patternStr) {
        this(buildMap(directory, patternStr));
    }


    /**
     *  Initialize a library of the files having one of the given list of
     *  extensions in the given directory.
     *  The keys are the unqualified file names.
     *
     * @param directory  the directory containing image files for the library
     * @param extensions the filename extensions files must match
     */
    public RImageFileLibrary(File directory, String[] extensions) {
        this(buildMap(directory, makePattern(extensions)));
    }


    /**
     *  Retrieve an image from the library.
     *
     * @param key  the identifying key of the image file
     * @return     the {@link java.awt.image.RenderedImage} object for the
     *             given key.
     */
    public RenderedImage getImage(String key) {
        File file = (File)fileMap.get(key);
        ParameterBlock pb = new ParameterBlock();
        pb.add(file.getPath());
        return JAI.create("FileLoad", pb);
    }


    /** 
     *  Build a HashMap with the base file name (without directory) as
     *  key and the full path name as data.
     *
     */
    static private HashMap buildMap(File directory, String patternStr) {
        HashMap hMap = new HashMap();
        if (!directory.isDirectory()) {
        }
        File[] files = (patternStr == null) ?  directory.listFiles() :
                directory.listFiles(new PatternFileFilter(patternStr));
        
        for (int i=0; i<files.length; ++i) {
            hMap.put(files[i].getName(), files[i]);
        }
        return hMap;
    }


    /**
     *  Make a regular expression pattern from a list of filename extensions.
     */
    static private String makePattern(String[] extensions) {
        StringBuffer sb = new StringBuffer(extensions[0]);
        for (int i=1; i<extensions.length; ++i) {
            sb.append("|");
            sb.append(extensions[i]);
        }
        return  ".*\\.(" + sb + ")$";
    }

}
