package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.util.Arrays;

/**
 *  A wrapper for access to intensities data.  The data could be passed in
 *  or read from a data source.
 */
public class RawIntensitiesData {

    private final RawIntensitiesRecord[] normalizedData;
    private int numChannels;


    public RawIntensitiesData(IntensitiesReader rdr, int numChannels,
            int indexBase, float intensityScale, boolean coerceNumChannels)
                    throws DatException {
        this(rdr.getData(), numChannels, indexBase, intensityScale,
                coerceNumChannels);
    }


    public RawIntensitiesData(IntensitiesReader rdr, int numChannels,
            int indexBase, float intensityScale)
                    throws DatException {
        this(rdr, numChannels, indexBase, intensityScale, false);
    }


    public RawIntensitiesData(RawIntensitiesRecord[] rawData, int numChannels,
            int indexBase, float intensityScale, boolean coerceNumChannels)
                    throws DatException {
        this.numChannels = numChannels;
        if (coerceNumChannels) {
            for (int i=0; i<rawData.length; ++i) {
                rawData[i].coerceNumChannels(numChannels);
            }
        }

        scanRawData(rawData, indexBase, intensityScale);
        normalizedData = new RawIntensitiesRecord[rawData.length];
        for (int i=0; i<rawData.length; ++i) {
            normalizedData[i] =
                    rawData[i].getNormalized(indexBase, intensityScale);
        }

        // Sort the records; hierrarchy=(zoneIndex,metaRow,metaCol,row,col)

        Arrays.sort(normalizedData);
    }


    private void scanRawData(RawIntensitiesRecord[] rawData, int indexBase, 
            float intensityScale) throws DatException {
        boolean includesZoneIndex = (rawData[0].zoneIndex != -1);
        boolean includesMetacoords = (rawData[0].metacol != -1);

        for (int i=0; i<rawData.length; ++i) {
            if (includesZoneIndex && rawData[i].zoneIndex < indexBase) {
                throw new DatException("zoneIndex less than index base",
                        null);
            }
            if (!includesZoneIndex && rawData[i].zoneIndex != -1) {
                throw new DatException("zoneIndex not consistently included",
                       null);
            }

            if (includesMetacoords && (rawData[i].metacol < indexBase ||
                    rawData[i].metarow < indexBase)) {
                throw new DatException("meta coordinates less than index base",
                        null);
            }
            if (!includesMetacoords && (rawData[i].metacol != -1 ||
                    rawData[i].metarow != -1))  {
                throw new DatException("meta coordinates not consistently " +
                        "included", null);
            }

            if (rawData[i].intensities.length != numChannels) {
                throw new DatException("wrong number (" +
                        String.valueOf(rawData[i].intensities.length) +
                        ") of intensities", null);
            }
        }
    }


    public RawIntensitiesRecord[] getData() {
        return normalizedData;
    }

}
