package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.util.Arrays;

/**
 *  Encapsulates the raw intensities data for one feature as well as its
 *  location (indices) in the microarray grid.
 *  These objects are sortable by position on the microarray.
 */
public class RawIntensitiesRecord  implements Comparable {
    public int zoneIndex;
    public int metarow;
    public int metacol;
    public int row;
    public int col;
    public float[] intensities;

    public RawIntensitiesRecord() {
        zoneIndex = -1;
        metarow = -1;
        metacol = -1;
        row = -1;
        col = -1;
    }


    public RawIntensitiesRecord getNormalized(int indexBase,
            float intensityScale) throws DatException {
        if (indexBase == 0  && intensityScale == 1f) {
            return this;
        }

        RawIntensitiesRecord norm = new RawIntensitiesRecord();
        norm.zoneIndex = (zoneIndex == -1) ? -1 : zoneIndex - indexBase;
        norm.metarow = (metarow == -1) ? -1 : metarow - indexBase;
        norm.metacol = (metacol == -1) ? -1 : metacol - indexBase;

        if (norm.zoneIndex == -1 && norm.metarow == -1 && norm.metacol == -1) {
            norm.zoneIndex = indexBase;
        }

        if (row < indexBase) {
            throw new DatException("Row (" + String.valueOf(row) + ") is less"
                    + " than index base (" + String.valueOf(indexBase) + ")" );
        }
        if (col < indexBase) {
            throw new DatException("Col (" + String.valueOf(col) + ") is less"
                    + " than index base (" + String.valueOf(indexBase) + ")" );
        }
        norm.row = row - indexBase;
        norm.col = col - indexBase;

        if (intensityScale == 1f) {
            norm.intensities = intensities;
        }
        else {
            norm.intensities = new float[intensities.length];
            for (int i=0; i<intensities.length; ++i) {
                norm.intensities[i] = intensities[i] / intensityScale;
            }
        }
    
        return norm;
    }


    public void coerceNumChannels(int numChan) {
        if (intensities.length == numChan) {
            return;
        }

        float[] saveIntensities = intensities;
        intensities = new float[numChan];
        int copyLen = (numChan > saveIntensities.length) ?
                saveIntensities.length : numChan;
 
        for (int i=0; i<copyLen; ++i) {
            intensities[i] = saveIntensities[i];
        }
    }


    static public void sort(RawIntensitiesRecord[] recs) {
        Arrays.sort(recs);
    }


    static public int zoneCompare(RawIntensitiesRecord rec1,
            RawIntensitiesRecord rec2) {
        if (rec1.zoneIndex >=0 && rec2.zoneIndex >=0) {
            return compInt(rec1.zoneIndex, rec2.zoneIndex);
        }

        int diff = compInt(rec1.metarow, rec2.metarow);
        return (diff != 0) ? diff :
                compInt(rec1.metacol, rec2.metacol);
    }


    static int compInt(int i1, int i2) {
        return (i1 == i2) ? 0 : (i1 < i2) ? -1 : 1;
    }


    public int compareTo(Object o) {
        
        RawIntensitiesRecord comp = (RawIntensitiesRecord)o;

        if (zoneIndex != comp.zoneIndex) {
            return compInt(zoneIndex, comp.zoneIndex);
        }
     
        if (metarow != comp.metarow) {
            return compInt(metarow, comp.metarow);
        }

        if (metacol != comp.metacol) {
            return compInt(metacol, comp.metacol);
        }

        if (row != comp.row) {
            return compInt(row, comp.row);
        }

        return compInt(col, comp.col);
    }

}
