package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.image.renderable.ParameterBlock;
import java.io.File;

import javax.media.jai.JAI;
import javax.media.jai.RenderedOp;

/**
 * Image scanner class, used by {@link Scan} class (which calls this class'
 * {@link #saveImages} method).
 */
public class Scanner {

    private File outDir = new File( System.getProperty("user.dir") );

    private boolean includeCybertoryMark = true;

    private final float pixelSize;

    private int blurKernelSize;
    private float blurStdDev;

    private PointModifyFunction noiseFunction = null;

    private ScanSaveType[] scanSaveTypes;

    static {
        GBlurOpDesc.init();
        PointModifyFunctionOpDesc.init();
    }

 
    /**
     * Constructor
     *
     * @param scanSaveTypes  a list of scanner save types (e.g., file formats)
     * @param pixelSize  the size of one pixel in microns (resolution)
     */
    public Scanner(ScanSaveType[] scanSaveTypes, float pixelSize) {
        this.scanSaveTypes = scanSaveTypes;
        this.pixelSize = pixelSize;
    }


    public void setBlur(int kernelSize, float stdDev) {
        this.blurKernelSize = kernelSize;
        this.blurStdDev = stdDev;
    }


    public void setNoiseFunction(PointModifyFunction noiseFunction) {
        this.noiseFunction = noiseFunction;
    }


    public void setIncludeCybertoryMark(boolean includeCybertoryMark) {
        this.includeCybertoryMark = includeCybertoryMark;
    }


    public void saveImages(PendingRenderedOp microarray, String fileStem,
            float gain) {

        RenderedOp image = microarray.manifestResolution(pixelSize);
        int numBands = image.getSampleModel().getNumBands();
        ParameterBlock pb;

        // blur
        if (blurKernelSize > 0  &&  blurStdDev > 0f) {
            pb = new ParameterBlock();
            pb.addSource(image);
            pb.add(blurKernelSize);
            pb.add(blurStdDev);
            image = JAI.create("GBlur", pb);
        }

        // noise
        if (noiseFunction != null) {
            pb = new ParameterBlock();
            pb.addSource(image);
            pb.add(noiseFunction);
            image = JAI.create("ModifyFunction", pb);
        }

        // gain
        if (gain != 1f) {
            pb = new ParameterBlock();
            pb.addSource(image);
            pb.add( new double[] {(double)gain} );
            image = JAI.create("MultiplyConst", pb);
        }

/*
        // Cybertory Mark
        if (includeCybertoryMark) {
            RenderedImage cybImage = CybertoryMark.image;
            RenderingHints rhCyb = FloatSampleModel.getRenderingHints(0, 0, 
                    cybImage.getWidth(), cybImage.getHeight(), numBands);

            if (numBands > 1) {
                int[] bands = new int[numBands];
                Arrays.fill(bands, 0);
                pb = new ParameterBlock();
                pb.addSource(cybImage);
                pb.add(bands);
                cybImage = JAI.create("BandSelect", pb, rhCyb);
            }

            pb = new ParameterBlock();
            pb.addSource(image);
            pb.addSource(cybImage);
            image = JAI.create("Overlay", pb);
        }
*/

        for (int i=0; i<scanSaveTypes.length; ++i) {
            scanSaveTypes[i].createFile(image, fileStem);
        }
    }


    /**
     * Set the output directory for the tiff/jpeg images --
     * defaults to the user's current directory.
     *
     * @param outDir  The path of the directory where to save image files.
     */
    public void setOutputDir(File outDir) {
        this.outDir = outDir;
    }

}
