package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.geom.Point2D;
import java.awt.image.RenderedImage;
import java.awt.image.renderable.ParameterBlock;
import java.util.Random;

import javax.media.jai.InterpolationBilinear;
import javax.media.jai.JAI;
import javax.media.jai.RenderedOp;

/**
 * An implementation of {@link SpotterTip} that uses a constant image.
 */
public class SpotterTipConst implements SpotterTip {

    private final RenderedImage tipTexture;
    private final float physicalTextureSize;
    /** The physical size (average of the 2 dimensions) of the constant image 
        passed into the constructor for this class */

    private SpotterTipLocationEffects locationEffects = null;

    public SpotterTipConst(RenderedImage tipTexture) {
        this(tipTexture, 0f);
    }


    public SpotterTipConst(RenderedImage tipTexture, float physicalTextureSize)
    {
        this.physicalTextureSize = physicalTextureSize;
        float transX = (float)
                (0.5 * tipTexture.getWidth() + tipTexture.getMinX());
        float transY = (float)
                (0.5 * tipTexture.getHeight() + tipTexture.getMinY());

        ParameterBlock pb = new ParameterBlock();
        pb.addSource(tipTexture);
        pb.add(-transX);
        pb.add(-transY);
        pb.add(new InterpolationBilinear());
        this.tipTexture = JAI.create("translate", pb);
    }


    public void setLocationEffects(SpotterTipLocationEffects locationEffects) {
        this.locationEffects = locationEffects;
    }


    public Object getScaled(float pixelSize) {
        SpotterTipConst newTip =
                new SpotterTipConst(tipTexture, physicalTextureSize/pixelSize);

        if (locationEffects != null) {
            newTip.setLocationEffects((SpotterTipLocationEffects)
                    locationEffects.getScaled(pixelSize));
        }

        return newTip;
    }


    private Point2D getImageCenter(RenderedImage rim) {
        return new Point2D.Float(
                (float)rim.getMinX() + 0.5f * (float)rim.getWidth(),
                (float)rim.getMinY() + 0.5f * (float)rim.getHeight() );
    }


    private float textureDiagonalSize() {
        int x = tipTexture.getWidth();
        int y = tipTexture.getHeight();
        return (float)
                Math.sqrt((((double)x)*(double)x) + (((double)y)*(double)y));
    }


    public SpotterTipContact getSpotterTipContact(Point2D featureCellCenter,
            float featureSize) {

        Point2D featureCenter = (locationEffects == null) ? featureCellCenter :
                locationEffects.getModifiedCenter(featureCellCenter);

        RenderedOp rop = null;

        if (tipTexture != null) {
    
            // If the actual physical size (in microns) of the texture image was
            // was specified, then the physical size should be preserved with
            // the new final resolution.  Otherwise, the image should be scaled
            // to match the featureSize.
            float scaleFactor = ((physicalTextureSize > 0f) ?
                    physicalTextureSize : featureSize) / textureDiagonalSize();
    
            ParameterBlock pb = new ParameterBlock();
            pb.addSource(tipTexture);
            pb.add(scaleFactor);
            pb.add(scaleFactor);
            pb.add((float)featureCenter.getX());
            pb.add((float)featureCenter.getY());
            pb.add(new InterpolationBilinear());
            rop = JAI.create("Scale", pb);
        }

        return new SpotterTipContact(featureCenter, rop);
    }
}
