package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.lang.reflect.Constructor;
import java.util.HashMap;

import org.jdom.Element;

/**
 *  Encapsulation of the handling of external and forward cached objects
 *  for access while parsing the XML.
 */
public class XMLAltData {

    public final int numberOfChannels;

    private ExternalDataFactory[] extDataFactories;
    //private HashMap externalData;
    private HashMap objectCache = new HashMap();

    private static final String extPrefix = "ext_";
    private static final String extTag = "key";

    private static final String refPrefix = "ref_";
    private static final String refTag = "id";


    public XMLAltData(ExternalDataFactory[] extDataFactories) {
        this.extDataFactories = extDataFactories;
        this.numberOfChannels = 2;
    }


    public XMLAltData(ExternalDataFactory[] extDataFactories,
            int numberOfChannels) {
        this.extDataFactories = extDataFactories;
        this.numberOfChannels = numberOfChannels;
    }


    static public boolean isAltElement(Element elem) {
        String eName = elem.getName();
        if (eName.length() < 4) {
           return false;
        }
        String prefix = eName.substring(0,4);
        return (prefix.equals(extPrefix) || prefix.equals(refPrefix));
    }


    static public String getAltLabel(boolean forExternal, String parseLabel) {
        return (forExternal ? extPrefix : refPrefix) + parseLabel;
    }


    static public String getTypeName(boolean forExternal) {
        return forExternal ? "ExternalData" : "BackwardRef";
    }


    public void cacheObject(String id, Object obj) {
        objectCache.put(id, obj);
    }


    public Object parseAltElement(Element elem, Class type)
            throws XMLParseException {
        String prefix = elem.getName().substring(0,4);
        boolean isExt = prefix.equals(extPrefix);
        String tagName = isExt ? extTag : refTag;
        String tag = elem.getAttributeValue(tagName);
        if (tag == null) {
            throw new XMLParseException("Missing attribute \"" + tagName + "\"",
                    elem);
        }

        Class oType = type.isPrimitive() ? getNonPrimitive(type) : type;

        Object obj = null;
        if (isExt) {
            for (int i=0; obj == null && i<extDataFactories.length; ++i) {
                try {
                    obj = extDataFactories[i].getExternalData(tag, oType);
                }
                catch (ExternalDataException e) {
                    throw new XMLParseException("external data exception",
                            elem, e);
                }
            }
        }
        else {
            obj = objectCache.get(tag);
        }

        if (obj == null) {
            throw new XMLMissingAltData(isExt, elem, tag);
        }

        //  Make sure that the data is of the correct type.
        //  For external data, if it's not, try to parse
        //  (if available).
        Class objClass = obj.getClass();
        if (oType.isAssignableFrom(objClass)) {
            return obj;
        }

        if (isExt && obj instanceof String) {
            try {
                Constructor constructor = 
                        oType.getConstructor(new Class[] { String.class });
                return constructor.newInstance(new Object[] { obj });
            }
            catch (NoSuchMethodException e) {
                // continue to throw exception below
            }
            catch (Exception e) {
                throw new XMLParseException("Data format error", elem, e);
            }
        }

        throw new XMLParseException("External data: expecting type " +
                oType.getName() + "; got type " + obj.getClass().getName(),
                        elem);
    }


    static private Class getNonPrimitive(Class primClass) {
        return primClass.equals(int.class) ? Integer.class :
               primClass.equals(float.class) ? Float.class :
               primClass.equals(double.class) ? Double.class :
               primClass.equals(short.class) ? Short.class :
               primClass.equals(long.class) ? Long.class :
               primClass.equals(byte.class) ? Byte.class :
               primClass.equals(boolean.class) ? Boolean.class :
               null;
    }
    

    static public Element schemaTypedef(boolean forExternal) {
        Element tagAttr = forExternal ?
                XMLUtils.xsdAttribute(extTag, "xsd:token") :
                XMLUtils.xsdAttribute(refTag, "xsd:IDREF");
        tagAttr.setAttribute("use", "required");

        String comment = forExternal ?
                "A means of using unique tags to refer to data outside of " +
                "the XML." :
                "A means of refering to elements previously defined in the " +
                "XML.";

        tagAttr.addContent(XMLUtils.xsdDocumentation(comment));
        return XMLUtils.xsdComplexType(getTypeName(forExternal), tagAttr);
    }

}
