package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.io.PrintStream;
import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.Vector;

import org.jdom.DocType;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.filter.ElementFilter;
import org.jdom.filter.Filter;

/**
 * The base class for code associated with handled data types, 
 * including functionality for generating XML Schema and parsing
 * the corresponding XML data.
 *
 * There's also a lot of static code behind the methods 
 * {@link #schemaDocument} and {@link #parseXMLDocument}.
 *
 */
public abstract class XMLHandler implements Comparable {

    static private HashMap lookup = new HashMap();
    /** map of handlers keyed by class */

    static private HashMap shortLookup = new HashMap();
    /** map of handlers keyed by short name for class */

    protected final Class handledClass;
    /** The Class whose data is parsed by this handler */

    static public void handlersReport(PrintStream output) {
        output.println("report not implemented yet");
    }


    /**
     * Given a class (or interface), return its handled subclasses.
     * @param superClass  the class for whom the subclasses are desired
     * @return  an array of subclasses of the given class
     */
    static public Class[] getHandledClasses(Class superClass) {
        Set handledClasses = lookup.keySet();
        if (superClass == null) {
            return (Class[])handledClasses.toArray(
                    new Class[handledClasses.size()]);
        }
        Vector subClasses = new Vector();
        Iterator iter = handledClasses.iterator();
        while (iter.hasNext()) {
            Class cls = (Class)iter.next();
            if (superClass.isAssignableFrom(cls)) {
                subClasses.add(cls);
            }
        }
        return (Class[])subClasses.toArray(new Class[subClasses.size()]);
    }


    static protected XMLHandler[] getAllHandlers() {
        Collection values =  lookup.values();
        XMLHandler[] allHandlers = (XMLHandler[])
                values.toArray(new XMLHandler[values.size()]);
        Arrays.sort(allHandlers);

        return allHandlers;
    }


    static protected XMLHandler[] getHandlersDefs() {
        Iterator handlers = lookup.values().iterator();
        Vector vDefsHandlers = new Vector();
        while (handlers.hasNext()) {
            XMLHandler handler = (XMLHandler)handlers.next();
            if (handler.allowBackwardRef()) {
                vDefsHandlers.add(handler);
            }
        }
        
        XMLHandler[] defsHandlers = (XMLHandler[])
                vDefsHandlers.toArray(new XMLHandler[vDefsHandlers.size()]);
        Arrays.sort(defsHandlers);

        return defsHandlers;
    }


    static private Element schemaWithinDefs(String typeName) {
        Element extElem = XMLUtils.xsd("extension");
        extElem.setAttribute("base", typeName);
        Element idAttr = XMLUtils.xsdAttribute("id", "xsd:ID");
        idAttr.setAttribute("use", "required");
        extElem.addContent(idAttr);

        Element complexContent = XMLUtils.xsd("complexContent");
        complexContent.addContent(extElem);
        Element complexType = XMLUtils.xsd("complexType");
        complexType.addContent(complexContent);
        
        Element elem = XMLUtils.xsdElement(XMLUtils.toLowerFirstChar(typeName),
                complexType);
        //elem.setAttribute("minOccurs", "0");
        //elem.setAttribute("maxOccurs", "unbounded");
 
        return elem;
    }


    /**
     *
     *
     */
    static private Element schemaDefs() {
        Element choiceElem = XMLUtils.xsd("choice");
        XMLHandler[] defsHandlers = getHandlersDefs();
        for (int i=0; i<defsHandlers.length; ++i) {
            String typeName = defsHandlers[i].xmlTypeName();
            choiceElem.addContent(schemaWithinDefs(typeName));
        }
        choiceElem.setAttribute("minOccurs", "0");
        choiceElem.setAttribute("maxOccurs", "unbounded");

        return XMLUtils.xsdElement("defs", 
                XMLUtils.xsd("complexType", choiceElem));
    }


    /**
     * Return a JDOM {@link Document} for the entire schema of the XML for
     * generating images.  All of the {@link XMLHandler} instances must be
     * created before this will work.
     *
     * @return a {@link Document} for the image generation schema
     */
    static public Document schemaDocument()
            throws XMLUnhandledTypeException {
        Element schemaRoot = XMLUtils.xsdRoot();

        schemaRoot.addContent(XMLAltData.schemaTypedef(true));
        schemaRoot.addContent(XMLAltData.schemaTypedef(false));

        Element scanElem = XMLUtils.xsdElement("scan", "Scan");
        scanElem.setAttribute("maxOccurs", "unbounded");
        
        schemaRoot.addContent(
               XMLUtils.xsdElement("microarrayRequest", 
                 XMLUtils.xsdComplexType(null, 
                   XMLUtils.xsdGrouping("sequence", new Element[] {
                   schemaDefs(), scanElem } )
                 ) ));

        Element scan2Elem = XMLUtils.xsdElement("scan", "Scan");
        schemaRoot.addContent(scan2Elem);
             
        XMLHandler[] allHandlers = getAllHandlers();

        // Put all the type definitions first
        for (int i=0; i<allHandlers.length; ++i) {
            if (allHandlers[i].needsTypedef()) {
                Element typedefElem = allHandlers[i].schemaTypedef();
                String description = allHandlers[i].typeDescription();
                if (description != null) {
                    typedefElem.addContent(0,
                            XMLUtils.xsdDocumentation(description));
                }
                schemaRoot.addContent(typedefElem);
            }
        }

        // Do another pass for the handlers that were created on the fly
        // (for interfaces and superclasses).
        allHandlers = getAllHandlers();
        for (int i=0; i<allHandlers.length; ++i) {
            if (allHandlers[i] instanceof XMLHandlerSuper &&
                    allHandlers[i].needsTypedef()) {
                schemaRoot.addContent(allHandlers[i].schemaTypedef());
            }
        }

        return new Document(schemaRoot, new DocType("Schema"));
    }


    /**
     * Parse the entire XML document requesting microarray image scans.
     *
     * @return an array of one or more {@link Scan} objects which can then
     *          be rendered (and saved to files).
     */
    static public Scan[] parseXMLDocument(Document doc,
            ExternalDataFactory[] extDataFactories)
            throws XMLParseException, XMLUnhandledTypeException {
        Element rootElem = doc.getRootElement();

        Element requestElem = rootElem.getChild("microarrayRequest");
        if (requestElem == null) {
            requestElem = rootElem;
        }

        return parseRequest(requestElem, new XMLAltData(extDataFactories));
    }

    
    /**
     *
     *
     */
    static public Scan[] parseRequest(Element parentElem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException {

        Element defsElem = parentElem.getChild("defs");
        if (defsElem != null) {
            Iterator chIter = defsElem.getChildren().iterator();
            while (chIter.hasNext()) {
                Element child = (Element)chIter.next();
                String label = XMLUtils.toUpperFirstChar(child.getName());

                XMLHandler handler = (XMLHandler)shortLookup.get(label);
                if (handler == null) {
                    throw new XMLParseException("Unexpected element", child);
                }
                Object obj = handler.parseElement(child, altData);
                String id = child.getAttributeValue("id");
                if (id == null) {
                    throw new XMLParseException("Missing \"id\" attribute",
                             child);
                }
                altData.cacheObject(id, obj);
            }
        }

        return (Scan[]) getHandler(Scan[].class).parseChildren(parentElem,
                    new ElementFilter("scan"), altData, false);
    }


    /**
     * Return the handler instance associated with a handled data type.
     *
     * @param cls the Class of interest
     * @return The XMLHandler instance associated with class of interest
     */
    static protected XMLHandler getHandler(Class cls)
            throws XMLUnhandledTypeException {
        XMLHandler handler = (XMLHandler)lookup.get(cls);
        if (handler != null)
            return handler;

        if (XMLHandlerAtomic.canInstantiate(cls)) {
            return new XMLHandlerAtomic(cls);
        }

        if (cls.isArray()) {
            handler = getHandler(cls.getComponentType());
            if (handler != null && handler.needsTypedef()) {
                return new XMLHandlerArray(cls, handler);
            }
        }

        if (XMLHandlerSuper.canInstantiate(cls)) {
            return new XMLHandlerSuper(cls);
        }

        return null;
    }


    // ********   END STATIC CODE ***********

    protected abstract boolean allowMultiple();
    /** Allow multiple elements for this data type ? */

    protected abstract boolean allowBackwardRef();
    /** Allow data in the "defs" section, with later backward references ? */

    protected abstract boolean needsTypedef();
    /** Does a schema typedef need to be generated for this data type ? */



    /**
     *  Constructor (never accessed by user code since abstract).
     */
    protected XMLHandler(Class handledClass) {
        this.handledClass = handledClass;
        lookup.put(handledClass, this);

        String name = handledClass.getName();
        String shortName = name.substring( name.lastIndexOf('.')+1 );
        shortLookup.put(shortName, this);
    }


    public String typeDescription() {
        return null;
    }


    /**
     * Generate Schema type definition for the class (using JDOM).
     *
     * @return a JDOM Element with schema for type
     */
    protected Element schemaTypedef() 
            throws XMLUnhandledTypeException {
        return null;
    }


    /**
     * Create the element for either "ext_*" or "ref_*" with given parse label.
     * @param parseLabel 
     * @param forExternal  indicates this should be an external data element
     *                     rather than a backward ref element.
     */
    protected static Element alternateElement(String parseLabel,
            boolean forExternal) {
        Element compElem = XMLUtils.xsd("complexType");
        Element idAttr = forExternal ?
                XMLUtils.xsdAttribute("key", "xsd:token") :
                XMLUtils.xsdAttribute("id", "xsd:IDREF");
        idAttr.setAttribute("usage", "required");
        compElem.addContent(idAttr);

        String name = (forExternal ? "ext_" : "ref_") + parseLabel;
        return XMLUtils.xsdElement(name, compElem);
    }


    protected Element schemaInContext(boolean required,
            boolean allowExternalData) throws XMLUnhandledTypeException {
        return schemaInContext(required, allowExternalData, null);
    }


    /**
     *  Return an element for a particular context
     *  
     * @param allowExternalData indicates whether to add an "externalData"
     *                          element.
     * @return an element for this data type
     */
    protected Element schemaInContext(boolean required,
            boolean allowExternalData, String description)
            throws XMLUnhandledTypeException {

        return schemaInContext(XMLUtils.toLowerFirstChar(xmlTypeName()),
            required, allowExternalData);
    }


    protected Element schemaInContext(String parseLabel, boolean required,
            boolean allowExternalData) throws XMLUnhandledTypeException {
        return schemaInContext(parseLabel, required,
            allowExternalData, null);
    }


    /**
     *  Return an element for a particular context
     *  
     * @param allowExternalData indicates whether to add an
     *                  "externalData" element.
     * @return an element for this data type
     */
    protected Element schemaInContext(String parseLabel, boolean required,
            boolean allowExternalData, String description)
            throws XMLUnhandledTypeException {
        if (parseLabel == null) {
            parseLabel = XMLUtils.toLowerFirstChar(xmlTypeName());
        }

        Element mainElem = XMLUtils.xsdElement(parseLabel);
        if (description != null) {
            mainElem.addContent(XMLUtils.xsdDocumentation(description));
        }
        String typeName = xmlTypeName();
        if (typeName != null) {
            mainElem.setAttribute("type", typeName);
        }
        else {
            mainElem.addContent(schemaTypedef());
        }

        Element topElem = mainElem;
        if (allowExternalData || allowBackwardRef()) {
            topElem = XMLUtils.xsd("choice");
            topElem.addContent(mainElem);
            if (allowExternalData) {
                topElem.addContent(XMLUtils.xsdElement(
                        XMLAltData.getAltLabel(true, parseLabel),
                        XMLAltData.getTypeName(true) ));
            }
            if (allowBackwardRef()) {
                topElem.addContent(XMLUtils.xsdElement(
                        XMLAltData.getAltLabel(false, parseLabel),
                        XMLAltData.getTypeName(false) ));
            }
        }
        if (!required) {
            topElem.setAttribute("minOccurs", "0");
        }
        if (allowMultiple()) {
            topElem.setAttribute("maxOccurs", "unbounded");
        }

        return topElem;
    }

    
    private Object parseSingleElement(Element elem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException {

        if (XMLAltData.isAltElement(elem)) {
            Class cls = handledClass.getComponentType();
            if (cls == null) {
                cls = handledClass;
            }
            return altData.parseAltElement(elem, cls);
        }

        else {
            return parseElement(elem, altData);
        }
    }


    /**
     * Parse zero or more children of the same data type.
     * Given the parent element below which the data resides for the
     * field of interest, parse it and return and object of the 
     * appropriate type for the handler.
     * If the data is absent, then either a null is returned or an
     * an exception is thrown, depending upon the 'required' field.
     *
     * @param parentElem  the JDOM element under which the data should be found.
     * @param childFilter a JDOM element filter object applied on the children
     * @param altData   Alternate source of data: external and cached data
     * @param required  indicates if it's not OK for the data to be absent
     * @return an object of the appropriate type (or null if data absent)
     */
    protected Object parseChildren(Element parentElem, Filter childFilter, 
            XMLAltData altData, boolean required)
            throws XMLParseException, XMLUnhandledTypeException {
        List childrenElems = extractChildren(parentElem, childFilter, required);

        if (!allowMultiple()) {
            if (childrenElems.size() == 0) {
                return null;
            }
            
            if (childrenElems.size() > 1) {
                throw new XMLParseException("Expecting only one child",
                        parentElem);
            }
            Element elem = (Element)childrenElems.get(0);
            return parseSingleElement(elem, altData);
        }
   
        Class atomicClass = handledClass.getComponentType();

        Object childrenArray =
                Array.newInstance(atomicClass, childrenElems.size());

        Iterator iter = childrenElems.iterator();
        int i=0;
        while (iter.hasNext()) {
            Element child = (Element)iter.next();
            String name = child.getName();
            Object obj = parseSingleElement(child, altData);
            Array.set(childrenArray, i++, obj);
        }
        return childrenArray;
    }


    /**
     *
     */
    protected List extractChildren(Element parentElem,
            Filter childFilter, boolean required) throws XMLParseException {
    
        List children = parentElem.getContent(childFilter);
        int numChildren = children.size();
        if (numChildren == 0) {
            if (required) {
                String options = "";
                if (childFilter instanceof XMLChildFilter) {
                    XMLChildFilter cFilt = (XMLChildFilter)childFilter;
                    options = cFilt.labelList();
                }
                throw new XMLParseException("Missing required element: " +
                        options, parentElem);
            }
        }
        else if (numChildren > 1 && !allowMultiple()) {
            throw new XMLParseException("Multiple occurrences not allowed",
                    null);
        }

        return children;
    }


    /**
     * Given that the element already extracted from it's context, parse
     * the data (whose type should match that of the handler).
     *
     * @param elem  the JDOM element containing the data
     * @param altData  the background data context for parsing
     * @return an object of the appropriate type
     */
    protected abstract Object parseElement(Element elem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException;


    /** 
     * Return the Class object for the data type handled by this wrapper.
     *
     * @return handled Class 
     */
    public Class handledClass() {
        return handledClass;
    }


    /**
     * Return the label to be used in the name of schema type defs.
     *
     * @return schema type label (or null if no data type name is to be used)
     */
    public abstract String xmlTypeName();


    public XMLChildFilter getXMLChildFilter(boolean allowExternalData) {
        return getXMLChildFilter(XMLUtils.toLowerFirstChar(xmlTypeName()),
                allowExternalData);
    }


    public XMLChildFilter getXMLChildFilter(String parseLabel,
            boolean allowExternalData) {
        return new XMLChildFilter(parseLabel, allowExternalData,
                allowBackwardRef());
    }


    /** To implement Comparable */
    public int compareTo(Object o) {
        String type1 = xmlTypeName();
        String type2 = ((XMLHandler)o).xmlTypeName();

        if (type1 == null) {
            return (type2 == null) ? 0 : -1;
        }
        else if (type2 == null) {
            return 1;
        }
        return type1.compareTo(type2);
    }

}
