package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.lang.reflect.Array;
import java.lang.reflect.Constructor;
import java.util.HashMap;

import org.jdom.Element;

/**
 *  One class that can serve for XMLHandler instances for all of:
 *  primitive types (int, float, boolean, etc.) and their object equivalents
 *  (Integer, Float, ... Boolean) and String as well as arrays of
 *  any of those things.
 */
public class XMLHandlerAtomic extends XMLHandler {

    private final int arrayDim; // num of array dimensions of type (0=not array)
    private final Class baseType;     // the component type (if array) or type
    private final Constructor objectConstructor;
    private final PrimitiveArrayer primitiveArrayer;


    static private HashMap primitiveMap = new HashMap();
    static {
        primitiveMap.put(byte.class, Byte.class);
        primitiveMap.put(short.class, Short.class);
        primitiveMap.put(int.class, Integer.class);
        primitiveMap.put(long.class, Long.class);
        primitiveMap.put(float.class, Float.class);
        primitiveMap.put(double.class, Double.class);
        primitiveMap.put(boolean.class, Boolean.class);
    }


    static private HashMap typeMap = new HashMap();
    static {
        typeMap.put(Byte.class,    "xsd:byte");
        typeMap.put(Short.class,   "xsd:short");
        typeMap.put(Integer.class, "xsd:integer");
        typeMap.put(Long.class,    "xsd:long");
        typeMap.put(Float.class,   "xsd:float");
        typeMap.put(Double.class,  "xsd:double");
        typeMap.put(Boolean.class, "xsd:boolean");
        typeMap.put(String.class,  "xsd:string");
        typeMap.put(Number.class,  "xsd:float");
    }


    static private boolean baseTypeSupported(Class cls) {
        return (cls.isPrimitive() || typeMap.containsKey(cls));
    }


    static protected boolean canInstantiate(Class cls) {
        if (baseTypeSupported(cls))
            return true;

        if (cls.isArray()) {
            Class comp = cls.getComponentType();
            if (baseTypeSupported(comp))
                return true;

            if (comp.isArray())
                return baseTypeSupported(comp.getComponentType());
        }

        return false;
    }


    public XMLHandlerAtomic(Class dataType) throws XMLUnhandledTypeException {
        super(dataType);
        if (!dataType.isArray()) {
            arrayDim = 0;
            baseType = dataType.equals(Number.class) ? Float.class : dataType;
        }
        else {
            Class compType = dataType.getComponentType();
            if (!compType.isArray()) {
                arrayDim = 1;
                baseType = compType.equals(Number.class) ? Float.class :
                        compType;
            }
            else {
                arrayDim = 2;
                baseType = compType.getComponentType();
            }
        }

        if (!(baseType.isPrimitive() || typeMap.containsKey(baseType))) {
            throw new XMLUnhandledTypeException(dataType);
        }
        
        primitiveArrayer = (arrayDim > 0 && baseType.isPrimitive()) ? 
                PrimitiveArrayer.getArrayer(baseType) : null;


        Class objectClass = baseType.isPrimitive() ?
                (Class)primitiveMap.get(baseType) : baseType;

        try {
            objectConstructor =
                    objectClass.getConstructor(new Class[] { String.class });
        }
        catch (Exception e) {
            throw new RuntimeException("Program bug: missing constructor");
        }
    }


    /**
     * Generate Schema type definition for the class (using JDOM).
     * In this case, it's only schema lists used for array data types.
     *
     * @return  a JDOM Element with schema for type
     */
    public Element schemaTypedef() throws XMLUnhandledTypeException {
        // Atomic types that are not arrays use built-in types, so there
        // is no schema.
        if (arrayDim == 0) {
            throw new RuntimeException("Shouldn't get here");
        }

        // Generate list schema
        Element listElem = XMLUtils.xsd("simpleType");
        String typeLabel = baseType.isPrimitive() ? "xsd:" + baseType.getName() 
                : (String)typeMap.get(baseType);

        if (arrayDim == 1) {
            listElem.addContent( XMLUtils.xsdList(typeLabel) );
        }
        else if (arrayDim == 2) {

            Element separatorElem = XMLUtils.xsd("enumeration");
            separatorElem.setAttribute("value", "|");
            Element restrType = XMLUtils.xsd("restriction");
            restrType.setAttribute("base", "xsd:string");
            restrType.addContent(separatorElem);

            Element listElemType = XMLUtils.xsd("union");
            listElemType.setAttribute("memberTypes", typeLabel);
            listElemType.addContent(XMLUtils.xsdSimpleType(null, restrType));

            listElem.addContent(listElemType);
        }
        return listElem;
    }


    private Object parseToken(String text) throws Exception {
        return objectConstructor.newInstance( new Object[] {text} );
    }


    private Object parseList(String text) throws Exception {
        String[] tokens = text.split("\\s+");
        Class compType = baseType.isPrimitive() ?
                (Class)primitiveMap.get(baseType) : baseType;
        Object[] objs = (Object[])Array.newInstance(compType, tokens.length);
        
        for (int i=0; i<objs.length; ++i) {
            objs[i] = parseToken(tokens[i]);
        }

        return (primitiveArrayer != null) ? primitiveArrayer.genArray(objs) :
                objs;
    }


    private Object parseList2(String text) throws Exception {
        String[] groups = text.split("|");
        Class compType = baseType.getComponentType();
        Object array = Array.newInstance(compType, groups.length);
        Object[] objs = new Object[groups.length];
        
        for (int i=0; i<groups.length; ++i) {
            Object obj = parseList(groups[i]);
            Array.set(array, i, obj);
        }
        return objs;
    }


    protected Object parseElement(Element elem, XMLAltData altData)
            throws XMLParseException {
        String value = elem.getTextTrim();
        try {
            switch (arrayDim) {
                case 0: return parseToken(value);
                case 1: return parseList(value);
                case 2: return parseList2(value);
            }
            throw new RuntimeException("Shouldn't get here");
        }
        catch (Exception e) {
            throw new XMLParseException("token parsing exception", elem, e);
        }
    }


   protected boolean allowMultiple() { return false; }
    /** Allow multiple elements for this data type ? */

    protected boolean allowBackwardRef() { return false; }
    /** Allow data in the "defs" section, with later backward references ? */

    protected boolean needsTypedef() { return false; }
    /** Does a schema typedef need to be generated for this data type ? */



    /**
     *
     */
    public String xmlTypeName() {
        // There's no distinct type name for an array of atomic types;
        // instead, in context anonymous list definitions are used.
        if (arrayDim > 0)
            return null;

        return handledClass.isPrimitive() ? "xsd:" + handledClass.getName()
                : (String)typeMap.get(handledClass);
    }

}
