package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.jdom.Element;

/**
 *  Class for automatic generation of XML handlers for super classes of
 *  the custom handlers.
 */
public class XMLHandlerSuper extends XMLHandler {

    static private HashSet candidates = null;
    /** a list of all superclasses and interfaces implemented by
        the handled classes of existing handlers */

    private final HashMap subsMap; /** subs handlers keyed by parse label */

    private XMLChildFilter childFilter = null;

    public boolean needsTypedef() { return true; }
    public boolean allowBackwardRef() { return false; }
    public boolean allowMultiple() { return false; }


    public XMLHandlerSuper(Class cls) throws XMLUnhandledTypeException {
        super(cls);
        subsMap = new HashMap();
        Class[] subClasses = getHandledClasses(cls);
        for (int i=0; i<subClasses.length; ++i) {
            if (cls.isAssignableFrom(subClasses[i]) &&
                    !cls.equals(subClasses[i])) {
                XMLHandler handler = getHandler(subClasses[i]);
                String parseLabel =
                        XMLUtils.toLowerFirstChar(handler.xmlTypeName());
                subsMap.put(parseLabel, handler);
            }
        }
        Set pLabelsSet = subsMap.keySet();
        childFilter = new XMLChildFilter(
                (String[])pLabelsSet.toArray(new String[pLabelsSet.size()]),
                true, true);
    }


    /** Generate a list of interfaces/classes that could be instantiated
     *  as a super. 
     */
    static private void getCandidates() {
        candidates = new HashSet();
        Class[] handledClasses = getHandledClasses(null);
        for (int i=0; i<handledClasses.length; ++i) {
            Class[] interfaces = handledClasses[i].getInterfaces();
            for (int j=0; j<interfaces.length; ++j) {
                candidates.add(interfaces[j]);
            }
            Class supClass = handledClasses[i].getSuperclass();
            if (supClass != null && !supClass.equals(Object.class)) {
                candidates.add(supClass);
            }
        }
    }


    /**
     *
     *
     */
    static boolean canInstantiate(Class cls) {
        if (candidates == null) {
            getCandidates();
        }
        return candidates.contains(cls);
    }


    /**
     * Generate Schema type definition for the class (using JDOM).
     *
     * @return a JDOM Element with schema for type definition
     */
    public Element schemaTypedef() throws XMLUnhandledTypeException {
        Iterator subsIter = subsMap.values().iterator();
        
        Element choiceElem = XMLUtils.xsd("choice");
        while (subsIter.hasNext()) {
            XMLHandler handler = (XMLHandler)subsIter.next();
            Element item = handler.schemaInContext(true, false);
            //Element item = handler.schemaInContext(false, false);
            //item.setAttribute("maxOccurs", "unbounded");
            choiceElem.addContent(item);
        }

        return XMLUtils.xsdComplexType(xmlTypeName(), choiceElem);
    }


   /**
     * Given that the element already extracted from it's context, parse
     * the data (whose type should match that of the handler).
     *
     * @param elem  the JDOM element containing the data
     * @param altData  the background data context for parsing
     * @return an object of the appropriate type
     */
    protected Object parseElement(Element elem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException {

        List children = elem.getContent(childFilter);
        if (children.size() != 1) {
            throw new XMLParseException("Single element expected under " +
                    xmlTypeName() + " element", elem);
        }
        Element childElem = (Element)children.get(0);

        if (XMLAltData.isAltElement(childElem)) {
            return altData.parseAltElement(childElem, handledClass);
        }

        String childName = childElem.getName();
        XMLHandler childHandler = (XMLHandler)subsMap.get(childName);
        if (childHandler == null) {
            throw new XMLParseException("Unexpected element: " + childName,
                    childElem);
        }

        return childHandler.parseElement(childElem, altData);
    }


    /**
     * Return the label to be used in the name of schema type defs.
     *
     * @return schema type label (or null if no data type name is to be used)
     */
    public String xmlTypeName() {
        String name = handledClass().getName();
        int dotPos = name.lastIndexOf(".");
        return (dotPos == -1) ? name : name.substring(dotPos+1);
    }

}
