package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import org.jdom.Element;
import org.jdom.Namespace;

/**
 *  General utilities for formatting and parsing XML.
 */
public class XMLUtils {

    static public Namespace xsdNamespace =
            Namespace.getNamespace("xsd", "http://www.w3.org/2001/XMLSchema");

    static public Element xsdDocumentation(String docText) {
        Element docElem = new Element("documentation", xsdNamespace);
        docElem.setText(docText);
        Element annElem = new Element("annotation", xsdNamespace);
        annElem.addContent(docElem);
        return annElem;
    }


    static public Element xsdRoot() {
        Element root = new Element("schema", xsdNamespace);
        root.addNamespaceDeclaration(xsdNamespace);
        return root;
    }


    static public Element xsdElement(String name) {
        return xsdElement(name, (String)null);
    }


    static public Element xsdElement(String name, Element subElem,
            boolean required) {
        Element elem = new Element("element", xsdNamespace);
        elem.setAttribute("name", name);
        elem.addContent(subElem);
        
        if (!required) {
            elem.setAttribute("minOccurs", "0");
        }

        return elem;
    }


    static public Element xsdElement(String name, Element subElem) {
        return xsdElement(name, subElem, true);
    }



    static public Element xsdElement(String name, String type,
            boolean required) {
        Element elem = new Element("element", xsdNamespace);
        elem.setAttribute("name", name);
        if (type != null) {
            elem.setAttribute("type", type);
        }
        if (!required) {
            elem.setAttribute("minOccurs", "0");
        }

        return elem;
    }


    static public Element xsdElement(String name, String type) {
        return xsdElement(name, type, true);
    }


    static public Element xsdEmptyElement(String name) {
        return xsdEmptyElement(name, true);
    }


    static public Element xsdEmptyElement(String name, boolean required) {
        Element elem = new Element("element", xsdNamespace);
        elem.setAttribute("name", name);
        elem.addContent(XMLUtils.xsd("complexType"));
        if (!required) {
            elem.setAttribute("minOccurs", "0");
        }

        return elem;
    }



    static public Element xsdAttribute(String name, String type) {
        Element elem = new Element("attribute", xsdNamespace);
        elem.setAttribute("name", name);
        elem.setAttribute("type", type);
        return elem;
    }


    static public Element xsdAttribute(String name, String type,
            boolean required) {
        Element elem = xsdAttribute(name, type);
        if (required) {
            elem.setAttribute("use", "required");
        }
        return elem;
    }


    static public Element xsd(String elemName, Element subElem) {
        Element elem = new Element(elemName, xsdNamespace);
        elem.addContent(subElem);
        return elem;
    }


    static public Element xsd(String elemName) {
        return new Element(elemName, xsdNamespace);
    }


    static public Element xsdList(String itemType) {
        Element elem = new Element("list", xsdNamespace);
        elem.setAttribute("itemType", itemType);
        return elem;
    }


    static public Element xsdSimpleType(String name, Element subElem) {
        Element elem = new Element("simpleType", xsdNamespace);
        if (name != null) {
            elem.setAttribute("name", name);
        }
        if (subElem != null) {
            elem.addContent(subElem);
        }

        return elem;
    }


    static public Element xsdComplexType(String name, Element subElem) {
        Element elem = new Element("complexType", xsdNamespace);
        if (name != null) {
            elem.setAttribute("name", name);
        }
        if (subElem != null) {
            elem.addContent(subElem);
        }

        return elem;
    }


    static public Element xsdGrouping(String groupType, Element[] choices) {
        return xsdGrouping(groupType, choices, true);
    }


    static public Element xsdGrouping(String groupType, Element[] choices,
                boolean required) {
        Element elem = new Element(groupType, xsdNamespace);
        if (choices != null) {
            for (int i=0; i<choices.length; ++i) {
                elem.addContent(choices[i]);
            }
        }

        if (!required) {
            elem.setAttribute("minOccurs", "0");
        }

        return elem;
    }


    static public boolean getChildBoolean(Element elem, String name)
            throws XMLMissingElementException {
        String str = elem.getChildText(name);
        if (str == null) {
            throw new XMLMissingElementException(elem, name);
        }

        return (str.equalsIgnoreCase("true") || str.equalsIgnoreCase("yes"));
    }


    static public boolean getChildBoolean(Element elem, String name,
            boolean defaultVal) {
        boolean val = defaultVal;
        try {
            val = getChildBoolean(elem, name);
        }
        catch (XMLMissingElementException e) {
            val = defaultVal;
        }

        return val;
    }



    static public String getChildString(Element elem, String name)
            throws XMLMissingElementException {
        String val = elem.getChildText(name);
        if (val == null) {
            throw new XMLMissingElementException(elem, name);
        }

        return val;
    }


    static public String getChildString(Element elem, String name,
            String defaultVal) {
        String val = elem.getChildText(name);
        return (val != null) ? val : defaultVal;
    }


    static public int getChildInteger(Element elem, String name)
            throws XMLMissingElementException, XMLFormatException {
        Element childElem = elem.getChild(name);
        if (childElem == null) {
            throw new XMLMissingElementException(elem, name);
        }
        
        try {
            return Integer.parseInt(childElem.getTextTrim());
        }
        catch (NumberFormatException e) {
            throw new XMLFormatException(childElem, e);
        }
    }


    static public int getChildInteger(Element elem, String name, int defaultVal)
            throws XMLFormatException {
        int val = defaultVal;
        try {
            val = getChildInteger(elem, name);
        }
        catch (XMLMissingElementException e) {
            val = defaultVal;
        }

        return val;
    }


    static public float getChildFloat(Element elem, String name)
            throws XMLMissingElementException, XMLFormatException {
        Element childElem = elem.getChild(name);
        if (childElem == null) {
            throw new XMLMissingElementException(elem, name);
        }

        try {
            return Float.parseFloat(childElem.getTextTrim());
        }
        catch (NumberFormatException e) {
            throw new XMLFormatException(childElem, e);
        }
    }


    static public float getChildFloat(Element elem, String name,
            float defaultVal) throws XMLFormatException {
        float val = defaultVal;
        try {
            val = getChildFloat(elem, name);
        }
        catch (XMLMissingElementException e) {
            val = defaultVal;
        }

        return val;
    }


    public static String toUpperFirstChar(String inStr) {
        return inStr.substring(0,1).toUpperCase() + inStr.substring(1);
    }


    public static String toLowerFirstChar(String inStr) {
        return inStr.substring(0,1).toLowerCase() + inStr.substring(1);
    }

}

