package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.Point;
import java.util.Iterator;
import java.util.Vector;

import org.jdom.Element;

/**
 *  XML wrapper for GridFeaturesIntensities.
 *
 *  For microarrays where all zones are the same size, located on a regular
 *  grid (evenly spaced metarows and metacols).
 *  Also, the spacing between rows and columns (within zones) is consistent
 *  amongst all the zones.
 */
public class X_GridFeaturesIntensities extends XMLHandlerStd {

    public X_GridFeaturesIntensities() throws XMLUnhandledTypeException {
        super(GridFeaturesIntensities.class, true);
    }


    public String typeDescription() {

        String shortName = handledClass.getName();
        int dotPos = shortName.lastIndexOf(".");
        if (dotPos >= 0) {
            shortName = shortName.substring(dotPos+1);
        }
        return shortName + 
         " is an implementation of " +
                "FeaturesIntensities for microarrays where all zones are " +
                "the same size, located on a regular grid (evenly spaced " +
                "metarows and metacols).  Also, the spacing between rows " +
                "and columns (within zones) is consistent amongst all the " +
                "zones.";
    }


    private XMLChildFilter riFilter = null;
    private XMLChildFilter nmcFilter = null;


    /**
     *  From somewhere, we have to get the number of metarows,metarows as well
     *  as the number of rows,cols.  This information could be explicit or 
     *  it could be inferred from the raw data.
     */
    public Element schemaTypedef() throws XMLUnhandledTypeException {

        return XMLUtils.xsdComplexType(xmlTypeName(), 
                XMLUtils.xsdGrouping("sequence", new Element[] {
                  XMLUtils.xsdGrouping("sequence", new Element[] {
                    XMLUtils.xsdElement("numRows", "xsd:int", false),
                    XMLUtils.xsdElement("numCols", "xsd:int", false) } ),
                  XMLUtils.xsdGrouping("sequence", new Element[] {
                    XMLUtils.xsdElement("numMetarows", "xsd:int", false),
                    getHandler(Integer.class).schemaInContext("numMetacols",
                      true, true) }, false ),
                  getHandler(RawIntensitiesData.class).schemaInContext(
                          true, false)
                } ) );
    }


    /**
     *
     */
    private float[][][] handleZone(Vector forZone,
            int numRows, int numCols,
            int maxRow, int maxCol, int numChannels, Element elem)
                    throws XMLParseException {

        if (numRows > 0) {
            if (maxRow >= numRows) {
                throw new XMLParseException("row (" + String.valueOf(maxRow)
                        + ") exceeds specified numRows (" +
                        String.valueOf(numRows) + ")", elem);
            }
        }
        else {
            numRows = maxRow + 1;
        }
    
        if (numCols > 0) {
            if (maxCol >= numCols) {
                throw new XMLParseException("col (" + String.valueOf(maxCol)
                        + ") exceeds specified numCols (" +
                        String.valueOf(numCols) + ")", elem);
            }
        }
        else {
            numCols = maxCol + 1;
        }


        float[][][] zoneData = new float[numRows][numCols][];
        Iterator iter = forZone.iterator();
        while (iter.hasNext()) {
            RawIntensitiesRecord rec = (RawIntensitiesRecord)iter.next();
            zoneData[rec.row][rec.col] = rec.intensities;
        }
        return zoneData;
    }


    /**
     *
     */
    private Point[] deriveMetaCoords(int numZones, int numMetacols) {
        Point[] map = new Point[numZones];
        if (numZones == 1) {
            numMetacols = 1;
        }
        for (int i=0; i<numZones; ++i) {
            map[i] = new Point(i%numMetacols, i/numMetacols);
        }
        return map;
    }


    /**
     *
     */
    public Object parseElement(Element elem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException {

        if (riFilter == null) {
            riFilter = getHandler(RawIntensitiesData.class).getXMLChildFilter(
                    false);
        }
        RawIntensitiesData rawData = (RawIntensitiesData)
                getHandler(RawIntensitiesData.class).parseChildren(elem,
                        riFilter, altData, true);

        RawIntensitiesRecord[] rawRecs = rawData.getData();

        boolean includesMetaCoords =
                (rawRecs[0].metarow >= 0 && rawRecs[0].metacol >= 0);

        // The grid dimensions might be explicit in the XML
        int numRows = XMLUtils.getChildInteger(elem, "numRows", 0);
        int numCols = XMLUtils.getChildInteger(elem, "numCols", 0);
        int numMetarows = XMLUtils.getChildInteger(elem, "numMetarows", 0);

        int numMetacols = 0;
        if (nmcFilter == null) {
            nmcFilter = getHandler(Integer.class).getXMLChildFilter(
                    "numMetacols", true);
        }
        Integer nmcInt = (Integer)
                getHandler(Integer.class).parseChildren(elem, nmcFilter,
                        altData, false);
        if (nmcInt != null) {
            numMetacols = nmcInt.intValue();
        }


        if (numRows < 0 || numCols < 0 || numMetarows < 0 || numMetacols < 0) {
            throw new XMLParseException("All coordinates must be positive",
                    elem);
        }

        Vector vMetaCoords = null;
        if (includesMetaCoords) {
            vMetaCoords = new Vector();
        }

        Vector vAllZonesData = new Vector();
        Vector forZone = new Vector();
        forZone.add(rawRecs[0]);
        int maxRow = rawRecs[0].row;
        int maxCol = rawRecs[0].col;
        int globMaxRow = 0;
        int globMaxCol = 0;
        int i=1;
        for (; i<rawRecs.length; ++i) {
            if (RawIntensitiesRecord.zoneCompare(rawRecs[i-1], rawRecs[i])
                    != 0) {

                vAllZonesData.add(handleZone(forZone, numRows, numCols,
                        maxRow, maxCol,
                        altData.numberOfChannels, elem));
                if (includesMetaCoords) {
                    vMetaCoords.add(new Point(rawRecs[i-1].metacol,
                            rawRecs[i-1].metarow));
                }

                if (maxRow > globMaxRow) {
                    globMaxRow = maxRow;
                }
                if (maxCol > globMaxCol) {
                    globMaxCol = maxCol;
                }

                forZone = new Vector();
                maxRow = 0;
                maxCol = 0;
            }
            if (rawRecs[i].row > maxRow) {
                maxRow = rawRecs[i].row;
            }
            if (rawRecs[i].col > maxCol) {
                maxCol = rawRecs[i].col;
            }
            forZone.add(rawRecs[i]);
        }
        vAllZonesData.add(handleZone(forZone, numRows, numCols,
                maxRow, maxCol, altData.numberOfChannels, elem));

        if (maxRow > globMaxRow) {
            globMaxRow = maxRow;
        }
        if (maxCol > globMaxCol) {
            globMaxCol = maxCol;
        }

        if (numRows <= 0) {
            numRows = globMaxRow + 1;
        }
        else if (globMaxRow > numRows) {
            throw new XMLParseException("Row in data (" +
                    String.valueOf(globMaxRow) + ") beyond specified number "
                    + "of rows (" + String.valueOf(numRows) + ")", elem);
        }

        if (numCols <= 0) {
            numCols = globMaxCol + 1;
        }
        else if (globMaxCol > numCols) {
            throw new XMLParseException("Col in data (" +
                    String.valueOf(globMaxCol) + ") beyond specified number "
                    + "of cols (" + String.valueOf(numCols) + ")", elem);
        }

        Point[] zoneMetaCoords;

        if (includesMetaCoords) {
            vMetaCoords.add(new Point(rawRecs[i-1].metacol,
                    rawRecs[i-1].metarow));

            if (numMetarows > 0 && rawRecs[i-1].metarow >= numMetarows) {
                throw new XMLParseException("metarow (" +
                        String.valueOf(rawRecs[i-1].metarow) +
                        ") violates specified numMetarows (" + 
                        String.valueOf(numMetarows) + ")", elem);
            }

            if (numMetacols > 0) {
                Iterator iter = vMetaCoords.iterator();
                while (iter.hasNext()) {
                    int metacol = ((Point)iter.next()).x;
                    if (metacol >= numMetacols) {
                        throw new XMLParseException("metacol (" +
                                String.valueOf(rawRecs[i-1].metacol) +
                                ") violates specified numMetacols (" + 
                                String.valueOf(numMetacols) + ")", elem);
                    }
                }
            }
            zoneMetaCoords = (Point[])
                    vMetaCoords.toArray(new Point[vMetaCoords.size()]);
        }

        else {
            if (numMetacols <= 0 && vAllZonesData.size() > 1) {
                throw new XMLParseException("Either data file must contain " +
                        "metarow/metacol or XML must contain numMetacols",
                        elem);
            }

            zoneMetaCoords =
                    deriveMetaCoords(vAllZonesData.size(), numMetacols);
        }

        if (numMetarows <= 0) {
            numMetarows  = zoneMetaCoords[zoneMetaCoords.length-1].y+1;
        }

        if (numMetacols <= 0) {
            for (int z=0; z<zoneMetaCoords.length; ++z) {
               if (zoneMetaCoords[z].x + 1 > numMetacols) {
                   numMetacols = zoneMetaCoords[z].x + 1;
               }
            }
        }
        
        float[][][][] intensities = (float[][][][]) vAllZonesData.toArray(
                new float[vAllZonesData.size()][][][] );

        return new GridFeaturesIntensities(intensities, 
            numMetarows, numMetacols, numRows, numCols, zoneMetaCoords);
    }
}
