package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.io.File;
import java.io.FileReader;
import java.util.regex.Pattern;

import org.jdom.Attribute;
import org.jdom.DataConversionException;
import org.jdom.Element;
import org.jdom.filter.ElementFilter;

/**
 *  XML wrapper for IntensitiesFileReader.
 */
public class X_IntensitiesFileReader extends XMLHandlerStd {

    public X_IntensitiesFileReader() throws XMLUnhandledTypeException {
        super(IntensitiesFileReader.class);
    }


    public String typeDescription() {

        String shortName = handledClass.getName();
        int dotPos = shortName.lastIndexOf(".");
        if (dotPos >= 0) {
            shortName = shortName.substring(dotPos+1);
        }
        return shortName + 
         " a flexible class to read " +
                "feature intensities data from a flat file. " +
                "Input data includes: zone index and/or metarow/metacol, " +
                "row, column, and intensities values.";
    }


    private XMLChildFilter patternFilter = null;
    private XMLChildFilter colsFilter = null;


    static public int idFromString(String columnStr) {
        if      (columnStr.equalsIgnoreCase("Ignore"))    
            return IntensitiesFileReader.ID_Ignore;
        else if (columnStr.equalsIgnoreCase("ZoneIndex")) 
            return IntensitiesFileReader.ID_ZoneIndex;
        else if (columnStr.equalsIgnoreCase("Metarow"))   
            return IntensitiesFileReader.ID_Metarow;
        else if (columnStr.equalsIgnoreCase("Metacol"))   
            return IntensitiesFileReader.ID_Metacol;
        else if (columnStr.equalsIgnoreCase("Row"))       
            return IntensitiesFileReader.ID_Row;
        else if (columnStr.equalsIgnoreCase("Col"))       
            return IntensitiesFileReader.ID_Col;
        else return -1;
    }


    static public Element enumElem(String value) {
        Element elem = XMLUtils.xsd("enumeration");
        elem.setAttribute("value", value);
        return elem;
    }


    public Element schemaTypedef() throws XMLUnhandledTypeException {

        // Set up the "columns" type:
        Element restrictElem = XMLUtils.xsd("restriction");
        restrictElem.setAttribute("base", "xsd:token");
        restrictElem.addContent(enumElem("Ignore"));
        restrictElem.addContent(enumElem("ZoneIndex"));
        restrictElem.addContent(enumElem("Metarow"));
        restrictElem.addContent(enumElem("Metacol"));
        restrictElem.addContent(enumElem("Row"));
        restrictElem.addContent(enumElem("Col"));

        Element colsTypeElem =
                XMLUtils.xsdSimpleType(null, 
                  XMLUtils.xsd("list",
                    XMLUtils.xsdSimpleType(null, 
                      restrictElem) ));

        Element colsOrPatElem = XMLUtils.xsdGrouping("choice", new Element[] {
                XMLUtils.xsdElement("columns", colsTypeElem, false),
                getHandler(String.class).schemaInContext("pattern", true,
                          true) } );
        colsOrPatElem.setAttribute("minOccurs", "0");

        Element typeElem = XMLUtils.xsdComplexType(xmlTypeName(), 
                XMLUtils.xsdGrouping("sequence", new Element[] {
                  getHandler(File.class).schemaInContext("inputFile", true,
                          false),
                  colsOrPatElem } ) );
        typeElem.addContent(
                XMLUtils.xsdAttribute("onlyOneZone", "xsd:boolean", false));

        return typeElem;
    }


    public Object parseElement(Element elem, XMLAltData altData)
            throws XMLParseException, XMLUnhandledTypeException {

        boolean onlyOneZone = false;
        Attribute attr = elem.getAttribute("onlyOneZone");
        if (attr != null) {
            try {
                onlyOneZone = attr.getBooleanValue();
            }
            catch (DataConversionException e) {
                throw new XMLParseException(
                        "Attribute (\"onlyOneZone\") format error",  elem, e);
            }
        }

        File inputFile = (File)
                getHandler(File.class).parseChildren(elem, 
                        new ElementFilter("inputFile"), altData, true);

        if (patternFilter == null) {
            patternFilter = getHandler(String.class).getXMLChildFilter(
                    "pattern", true);
        }
        String patternStr = (String)
                getHandler(String.class).parseChildren(elem, patternFilter,
                        altData, false);
        Pattern pattern = null;
        if (patternStr != null) {
            pattern = Pattern.compile(patternStr);
        }

        if (colsFilter == null) {
            colsFilter = new XMLChildFilter("columns", true, false);
        }
        String[] columns = (String[])
                getHandler(String[].class).parseChildren(elem, colsFilter,
                        altData, false);

        int[] columnIds = null;
        if (columns != null && columns.length > 0) {
            columnIds = new int[columns.length];
            for (int i=0; i<columns.length; ++i) {
                columnIds[i] = idFromString(columns[i]);
            }
        }

        try {
            return new IntensitiesFileReader(new FileReader(inputFile),
                    columnIds, pattern, onlyOneZone);
        }
        catch (DatMissingColumn e) {
            throw new XMLParseException("File data exception", elem, e);
        }
        catch (java.io.FileNotFoundException e) {
            throw new XMLParseException("File not found: " +
                    inputFile.getPath(), elem, e);
        }

    }

}
