package org.cybertory.MSS;

/*--------------------------------------------------------
|                   COPYRIGHT (c) 2005
|          Attotron Biotechnologies Corporation
|    Distributed under the GNU general public license:
|     http://www.fsf.org/licensing/licenses/gpl.txt
|
| Avaliable from: http://www.cybertory.org
|         Author: Bill Martin  <eco_bill@yahoo.com>
|      Funded by: NIH SBIR grant #R44 RR13645 02A2
--------------------------------------------------------*/
import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.RenderingHints;
import java.awt.Shape;
import java.awt.geom.Point2D;
import java.awt.image.BufferedImage;
import java.awt.image.RenderedImage;
import java.awt.image.SampleModel;
import java.awt.image.renderable.ParameterBlock;
import java.awt.image.renderable.RenderedImageFactory;
import java.util.Arrays;
import java.util.Vector;

import javax.media.jai.ImageFunction;
import javax.media.jai.ImageLayout;
import javax.media.jai.JAI;
import javax.media.jai.OperationDescriptor;
import javax.media.jai.OperationDescriptorImpl;
import javax.media.jai.OperationRegistry;
import javax.media.jai.ParameterBlockJAI;
import javax.media.jai.ParameterListDescriptor;
import javax.media.jai.ParameterListDescriptorImpl;
import javax.media.jai.RenderedOp;
import javax.media.jai.registry.RIFRegistry;


/**
 * A JAI operation for painting the features within a zone.
 *
 * This class is designed to be useable in with
 * {@link PendingRenderedOp} /  {@link PendingParameterBlockJAI}.
 */
public class ZoneFeaturesImageOpDesc  extends OperationDescriptorImpl
        implements RenderedImageFactory  {

    public static String OPERATION_NAME = "ZoneFeaturesImage";
    private static boolean isInitialized = false;

    public static void init() {
        if (isInitialized)
            return;

        isInitialized = true;
        OperationRegistry opReg =
                JAI.getDefaultInstance().getOperationRegistry();
        OperationDescriptor od = new ZoneFeaturesImageOpDesc();

        opReg.registerDescriptor(od);
        RIFRegistry.register(opReg, od.getName(), OPERATION_NAME,
                             (RenderedImageFactory)od);
    }


    public ZoneFeaturesImageOpDesc( ) {

        super(
                // resources
                new String[][] {
                     { "GlobalName",  OPERATION_NAME },
                     { "LocalName",  OPERATION_NAME },
                     { "Vendor",  "Attotron Biosensor Corporation" },
                     { "Description", 
                          "The features component image for a zone" },
                     { "DocURL",  "" },
                     { "Version",  "0.9" },
                     { "arg0Desc", "Zone index" },
                     { "arg1Desc", "Number of columns" },
                     { "arg2Desc", "Number of rows" },
                     { "arg3Desc", "Number of color channels" },
                     { "arg4Desc",
                             "Globally shared specifications of features" },
                     { "arg5Desc", "zone intensities" },
                     { "arg6Desc",
                             "zone tip, shape and texture characteristics" },
                },

                // supported modes
                new String[] { "Rendered" },

                // source names
                null,   // use default naming convention "source0", ...

                // source classes
                new Class[][] { { } },

                new ParameterListDescriptorImpl(
                    // descriptor ??
                    new Object(),

                    // paramNames
                    new String[] { "ZoneIndex", "NumCols", "NumRows",
                            "NumberOfChannels", "FeaturesSpecs",
                            "FeaturesIntensities", "ZoneCharacteristics" },

                    // paramClasses
                    new Class[]
                         { Integer.class, Integer.class, Integer.class,
                           Integer.class,
                           FeaturesSpecs.class, float[][][].class, 
                           ZoneCharacteristics.class },

                    // paramDefaults
                    new Object[] {
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            ParameterListDescriptor.NO_PARAMETER_DEFAULT,
                            null },

                    // validParamValues
                    null
                )
        );
    }


    /**
     * Method needed for implementation of {@link RenderedImageFactory} 
     */
    public RenderedImage create(final ParameterBlock pbIn,
            final RenderingHints rh)  {

        int zoneIndex = pbIn.getIntParameter(0);
        int numCols = pbIn.getIntParameter(1);
        int numRows = pbIn.getIntParameter(2);
        int numChannels = pbIn.getIntParameter(3);
        FeaturesSpecs featuresSpecs = (FeaturesSpecs)pbIn.getObjectParameter(4);
        float[][][] featuresIntensities =
                (float[][][])pbIn.getObjectParameter(5);
        ZoneCharacteristics zoneCh =
                (ZoneCharacteristics)pbIn.getObjectParameter(6);


        RenderingHints rHints = null;
        if (rh != null) {
            rHints = (RenderingHints)rh.clone();
        }

        ParameterBlock pb;

        int imageWidth =
                (int)((float)numCols * featuresSpecs.featureCellSizeX);
        int imageHeight =
                (int)((float)numRows * featuresSpecs.featureCellSizeY);
        SampleModel flSampleModel = new FloatSampleModel(imageWidth,
                imageHeight, numChannels);
        ImageLayout il = new ImageLayout(0, 0, imageWidth, imageHeight);
        il.setSampleModel(flSampleModel);
        if (rHints == null) {
            rHints = new RenderingHints(JAI.KEY_IMAGE_LAYOUT, il);
        }
        else {
            rHints.put(JAI.KEY_IMAGE_LAYOUT, il);
        }

        // Checkerboard of feature intensities
        pb = new ParameterBlock();
        pb.add(new CheckerboardFunction(featuresIntensities, numChannels));
        pb.add(imageWidth);
        pb.add(imageHeight);
        pb.add(1f/featuresSpecs.featureCellSizeX);
        pb.add(1f/featuresSpecs.featureCellSizeY);
        RenderedOp checkerboard = JAI.create("ImageFunction", pb, rHints);

        SpotterTipContact[][] tipContacts = null;
        FeatureShapeFactory[][] shapeFactories = null;
        FeatureTexture[][] featureTextures = null;
        if (zoneCh != null) {
            if (zoneCh.tip != null) {
                tipContacts = new SpotterTipContact[numRows][numCols];
            }
            if (zoneCh.featureCharacteristics != null) {
                shapeFactories = new FeatureShapeFactory[numRows][numCols];
                featureTextures = new FeatureTexture[numRows][numCols];
                for (int i=0; i<zoneCh.featureCharacteristics.length; ++i) {
                    FeatureCharacteristics fc =
                            zoneCh.featureCharacteristics[i];
                    shapeFactories[fc.row][fc.col] = fc.factory;
                    featureTextures[fc.row][fc.col] = fc.texture;
                }
            }
        }

        // Prepare to create an image mask for the features shapes
        Shape[][] featuresShapes = new Shape[numRows][numCols];

        // Iterate through the entire zone grid, handle each cell that has
        // a feature:
        //  1) Locate the center of the feature.
        //  2) Also instantiate the SpotterTipContact and use its location
        //     if applicable.
        //  3) Generate the feature shape.
        Point2D[][] featureCenter =  new Point2D[numRows][numCols];
        for (int row=0; row<numRows; ++row) {
            for (int col=0; col<numCols; ++col) {
                if (featuresIntensities[row][col] != null) {
                    FeatureShapeFactory shapeFactory = null;
                    if (shapeFactories != null) {
                        shapeFactory = shapeFactories[row][col];
                    }
                    if (shapeFactory == null) {
                        shapeFactory = featuresSpecs.defaultShapeFactory;
                    }

                    if (shapeFactory == null) {
                        throw new RuntimeException("Feature without factory");
                    }

                    Point2D fCenter = new Point2D.Float(
                            ((float)col + 0.5f)*featuresSpecs.featureCellSizeX,
                            ((float)row + 0.5f)*featuresSpecs.featureCellSizeY);
                    if (zoneCh != null && zoneCh.tip != null) {
                        tipContacts[row][col] = zoneCh.tip.getSpotterTipContact(
                                fCenter, featuresSpecs.featureSize);
                        fCenter = tipContacts[row][col].tipCenter;
                    }
                    featuresShapes[row][col] = shapeFactory.createFeatureShape(
                            featuresSpecs.featureSize, fCenter);
                    featureCenter[row][col] = fCenter;
                }
            }
        }

        BufferedImage featuresMaskImage =
                genFeaturesMask(imageWidth, imageHeight, featuresShapes);

        pb = new ParameterBlock();
        pb.addSource(checkerboard);
        pb.addSource(featuresMaskImage);
        RenderedOp featuresImage = JAI.create("Multiply", pb, rHints);

        if (tipContacts != null) {
            RenderedImage tipsTextureImage =
                    genTipsTextureImage(tipContacts, imageWidth, imageHeight);

            pb = new ParameterBlock();
            pb.addSource(featuresImage);
            pb.addSource(tipsTextureImage);
            featuresImage = JAI.create("Multiply", pb, rHints);
        }

        if (featuresSpecs.defaultTexture != null || featureTextures != null) {
            RenderedImage textureImage =
                    genTextureImage(featuresShapes,
                    featuresSpecs.defaultTexture, featureTextures,
                    imageWidth, imageHeight, numChannels);
            if (textureImage != null) {
                pb = new ParameterBlock();
                pb.addSource(featuresImage);
                pb.addSource(textureImage);
                featuresImage = JAI.create("Multiply", pb, rHints);
            }
        }

        return featuresImage;
    }


    private BufferedImage genFeaturesMask(
            int imageWidth, int imageHeight,
            Shape[][] featuresShapes) {
        BufferedImage bi = new BufferedImage(imageWidth, imageHeight,
                BufferedImage.TYPE_BYTE_BINARY);
        Graphics2D g = bi.createGraphics();
        g.setColor(Color.white);
        
        for (int row=0; row<featuresShapes.length; ++row) {
            for (int col=0; col<featuresShapes[row].length; ++col) {
                if (featuresShapes[row][col] != null) {
                     g.fill(featuresShapes[row][col]);
                }
            }
        }

        return bi;
    }


    private RenderedImage genTipsTextureImage(
            SpotterTipContact[][] tipContacts,
            int imageWidth, int imageHeight) {
        
        ParameterBlockJAI pbj = new ParameterBlockJAI("Mosaic");
        for (int row=0; row<tipContacts.length; ++row) {
            for (int col=0; col<tipContacts[row].length; ++col) {
                RenderedImage rim = tipContacts[row][col].markImage;
                if (rim != null) {
                    pbj.addSource(rim);
                }
            }
        }
        pbj.setParameter("sourceThreshold", new double[][] {{1e-50}} );
        pbj.setParameter("backgroundValues", new double[] {1.0});

        RenderingHints rh = FloatSampleModel.getRenderingHints(0, 0,
                imageWidth, imageHeight, 1);

        RenderedOp rop = JAI.create("Mosaic", pbj, rh);

        return rop;
    }


    private RenderedImage genTextureImage(Shape[][] featuresShapes,
            FeatureTexture defaultTexture, FeatureTexture[][] featureTextures,
            int imageWidth, int imageHeight, int numChannels) {

        Vector singleChannelImages = new Vector();
        Vector multiChannelImages = new Vector();
        
        ParameterBlock pb;
        ParameterBlockJAI pbjs = new ParameterBlockJAI("Mosaic");
        ParameterBlockJAI pbjm = new ParameterBlockJAI("Mosaic");
        boolean hasSingle = false;
        boolean hasMulti = false;

        for (int row=0; row<featuresShapes.length; ++row) {
            for (int col=0; col<featuresShapes[row].length; ++col) {
                if (featuresShapes[row][col] != null) {
                    FeatureTexture texture = (featureTextures != null &&
                            featureTextures[row][col] != null) ?
                            featureTextures[row][col] : defaultTexture;
                    if (texture != null) {
                        RenderedImage textureImage = texture.createTexture(
                                featuresShapes[row][col], numChannels);
                        int tNumChannels =
                                textureImage.getSampleModel().getNumBands();

                        if (tNumChannels == 1) {
                            pbjs.addSource(textureImage);
                            hasSingle = true;
                        }
                        else if (tNumChannels == numChannels) {
                            pbjm.addSource(textureImage);
                            hasMulti = true;
                        }
                    }
                }
            }
        }

        RenderedOp single = null;
        RenderedOp multi = null;

        if (hasSingle) {
            pbjs.setParameter("sourceThreshold", new double[][] {{1e-50}} );
            pbjs.setParameter("backgroundValues", new double[] {1.0});

            pb = new ParameterBlock();
            pb.add((float)imageWidth);
            pb.add((float)imageHeight);
            pb.add( new Float[] { new Float(1f) } );
            RenderedOp bg = JAI.create("constant", pb);

            pbjs.addSource(bg);

            RenderingHints rh = FloatSampleModel.getRenderingHints(0, 0,
                    imageWidth, imageHeight, 1);

            single = JAI.create("Mosaic", pbjs, rh);
        }

        if (hasMulti) {
            pbjm.setParameter("sourceThreshold", new double[][] {{1e-50}} );
            pbjm.setParameter("backgroundValues", new double[] {1.0});

            Float[] values = new Float[numChannels];
            Arrays.fill(values, new Float(1f));
            pb = new ParameterBlock();
            pb.add((float)imageWidth);
            pb.add((float)imageHeight);
            pb.add(values);
            RenderedOp bg = JAI.create("constant", pb);

            pbjs.addSource(bg);

            RenderingHints rh = FloatSampleModel.getRenderingHints(0, 0,
                    imageWidth, imageHeight, numChannels);

            multi = JAI.create("Mosaic", pbjm, rh);

            if (hasSingle) {
                pb = new ParameterBlock();
                pb.addSource(multi);
                pb.addSource(single);
                return JAI.create("Multiply", pb, rh);
            }
            else {
                return multi;
            }
        }

        return single;
    }



    /**
      Produces an image so that the number of pixels horizontally & vertically 
      matches the number of columns and rows of features, respectively.
      The image can be scaled to the desired size with parameters to the 
      {@link ImageFunction} JAI operation.
     */
    public class CheckerboardFunction  implements ImageFunction {
    
        private final float[][][] intensities;
        private final int numChannels;
    
        public CheckerboardFunction(float[][][] intensities, int numChannels) {
            this.intensities = intensities;
            this.numChannels = numChannels;
        }
    
    
        public void getElements(double startX, double startY,
                double deltaX, double deltaY, int countX, int countY,
                int element, double[] real, double[] imag) {
    
            int xInd, yInd;
            int i=0;
            for (int y=0; y<countY; ++y) {
                yInd = (int)Math.floor(startY + (double)y * deltaY);
                for (int x=0; x<countX; ++x) {
                    xInd = (int)Math.floor(startY + (double)x * deltaY);
                    float[] pixelIntensities = intensities[yInd][xInd];
                    real[i++] = (pixelIntensities == null) ? 0.0 : 
                            (double)pixelIntensities[element]/255.0;
                }
            }
        }
    
    
        public void getElements(float startX, float startY,
                float deltaX, float deltaY, int countX, int countY,
                int element, float[] real, float[] imag) {
    
            int xInd, yInd;
            int i=0;
            for (int y=0; y<countY; ++y) {
                yInd = (int)Math.floor(startY + (float)y * deltaY);
                for (int x=0; x<countX; ++x) {
                    xInd = (int)Math.floor(startX + (float)x * deltaX);
                    float[] pixelIntensities = intensities[yInd][xInd];
                    real[i++] = (pixelIntensities == null) ? 0f : 
                            pixelIntensities[element]/255f;
                }
            }
        }
    
    
        public int getNumElements() {
            return numChannels;
        }
    
    
        public boolean isComplex() {
            return false;
        }
    
    }

}
